"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Encoder = void 0;
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const BinaryType_1 = require("../../types/rga-binary/BinaryType");
const ClockEncoder_1 = require("../../../json-crdt-patch/codec/clock/ClockEncoder");
const ConstantType_1 = require("../../types/const/ConstantType");
const CrdtEncoder_1 = require("../../../json-crdt-patch/util/binary/CrdtEncoder");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const StringType_1 = require("../../types/rga-string/StringType");
const utf8_1 = require("../../../util/utf8");
const ValueType_1 = require("../../types/lww-value/ValueType");
class Encoder extends CrdtEncoder_1.CrdtEncoder {
    encode(model) {
        delete this.clockEncoder;
        delete this.time;
        this.literals = new Map();
        this.reset();
        const clock = model.clock;
        const isServerClock = clock.getSessionId() === 1;
        if (isServerClock) {
            this.time = clock.time;
            this.b1vuint56(true, clock.time);
        }
        else {
            model.advanceClocks();
            this.clockEncoder = new ClockEncoder_1.ClockEncoder(model.clock);
        }
        this.encodeLiteralsTable(model);
        this.encodeRoot(model.root);
        const data = this.flush();
        if (isServerClock)
            return data;
        this.encodeClockTable(data);
        delete this.clockEncoder;
        delete this.literals;
        return this.flush();
    }
    encodeLiteralsTable(model) {
        const literalFrequencies = new Map();
        for (const node of model.nodes.iterate()) {
            if (node instanceof ObjectType_1.ObjectType) {
                for (const key of node.latest.keys()) {
                    const count = literalFrequencies.get(key) || 0;
                    literalFrequencies.set(key, count + 1);
                }
            }
            else if (node instanceof StringType_1.StringType) {
                for (const chunk of node.chunks()) {
                    const str = chunk.str;
                    if (str) {
                        const count = literalFrequencies.get(str) || 0;
                        literalFrequencies.set(str, count + 1);
                    }
                }
            }
            else if (node instanceof ValueType_1.ValueType) {
                const value = node.value;
                switch (typeof value) {
                    case 'number':
                    case 'string': {
                        const count = literalFrequencies.get(value) || 0;
                        literalFrequencies.set(value, count + 1);
                    }
                }
            }
        }
        const literals = [];
        for (const [literal, frequency] of literalFrequencies.entries())
            if (frequency > 1)
                literals.push(literal);
        this.encodeArray(literals);
        for (let i = 0; i < literals.length; i++)
            this.literals.set(literals[i], i);
    }
    encodeClockTable(data) {
        const clockEncoder = this.clockEncoder;
        const length = clockEncoder.table.size;
        const dataSize = data.byteLength;
        this.uint8 = new Uint8Array(8 + 12 * length + dataSize);
        this.view = new DataView(this.uint8.buffer, this.uint8.byteOffset, this.uint8.byteLength);
        this.offset = 0;
        this.b1vuint56(false, length);
        for (const sid of clockEncoder.table.keys()) {
            const ts = clockEncoder.clock.clocks.get(sid);
            if (ts)
                this.uint53vuint39(sid, ts.time);
            else if (sid === clockEncoder.clock.getSessionId())
                this.uint53vuint39(sid, 0);
            else {
            }
        }
        this.buf(data, dataSize);
    }
    ts(ts) {
        if (this.clockEncoder) {
            const relativeId = this.clockEncoder.append(ts);
            this.id(relativeId.sessionIndex, relativeId.timeDiff);
        }
        else {
            const sessionId = ts.getSessionId();
            if (sessionId === 1) {
                this.b1vuint56(true, this.time - ts.time);
            }
            else {
                this.b1vuint56(false, sessionId);
                this.vuint57(ts.time);
            }
        }
    }
    encodeRoot(root) {
        this.ts(root.id);
        if (root.node)
            this.encodeNode(root.node);
    }
    encodeNode(node) {
        if (node instanceof ObjectType_1.ObjectType)
            return this.encodeObj(node);
        else if (node instanceof ArrayType_1.ArrayType)
            return this.encodeArr(node);
        else if (node instanceof StringType_1.StringType)
            return this.encodeStr(node);
        else if (node instanceof ValueType_1.ValueType)
            return this.encodeVal(node);
        else if (node instanceof ConstantType_1.ConstantType)
            return this.encodeConst(node);
        else if (node instanceof BinaryType_1.BinaryType)
            return this.encodeBin(node);
        throw new Error('UNKNOWN_NODE');
    }
    encodeObj(obj) {
        const length = obj.latest.size;
        this.encodeObjectHeader(length);
        this.ts(obj.id);
        for (const [key, chunk] of obj.latest.entries())
            this.encodeObjChunk(key, chunk);
    }
    encodeObjChunk(key, chunk) {
        this.ts(chunk.id);
        const length = (0, utf8_1.utf8Count)(key);
        const indexInLiteralsTable = this.literals.get(key);
        if (indexInLiteralsTable !== undefined) {
            this.b1vuint56(true, this.literals.get(key));
        }
        else {
            this.b1vuint56(false, length);
            this.encodeUtf8(key, length);
        }
        this.encodeNode(chunk.node);
    }
    encodeArr(obj) {
        const length = obj.size();
        this.encodeArrayHeader(length);
        this.ts(obj.id);
        for (const chunk of obj.chunks())
            this.encodeArrChunk(chunk);
    }
    encodeArrChunk(chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            const nodes = chunk.nodes;
            const length = nodes.length;
            this.b1vuint56(false, length);
            this.ts(chunk.id);
            for (let i = 0; i < length; i++)
                this.encodeNode(nodes[i]);
        }
    }
    encodeStr(obj) {
        const length = obj.size();
        this.encodeStringHeader(length);
        this.ts(obj.id);
        for (const chunk of obj.chunks())
            this.encodeStrChunk(chunk);
    }
    encodeStrChunk(chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            const text = chunk.str;
            const indexInLiteralsTable = this.literals.get(text);
            const useLiteralsTable = indexInLiteralsTable !== undefined;
            let length;
            this.b1vuint56(false, useLiteralsTable ? 0 : (length = (0, utf8_1.utf8Count)(text)));
            this.ts(chunk.id);
            if (useLiteralsTable) {
                this.vuint57(indexInLiteralsTable);
            }
            else {
                this.encodeUtf8(text, length);
            }
        }
    }
    encodeBin(obj) {
        const length = obj.size();
        this.encodeBinaryHeader(length);
        this.ts(obj.id);
        for (const chunk of obj.chunks())
            this.encodeBinChunk(chunk);
    }
    encodeBinChunk(chunk) {
        if (chunk.deleted) {
            this.b1vuint56(true, chunk.deleted);
            this.ts(chunk.id);
        }
        else {
            const data = chunk.buf;
            const length = data.byteLength;
            this.b1vuint56(false, length);
            this.ts(chunk.id);
            this.buf(data, length);
        }
    }
    encodeConst(obj) {
        const { value } = obj;
        switch (value) {
            case null:
            case false:
            case true:
                this.encodeAny(value);
                break;
            case undefined:
                this.u8(0xc1);
                break;
            default: {
                if (typeof value === 'number')
                    this.encodeNumber(value);
                else {
                    this.u8(0xd4);
                    this.encodeAny(value);
                }
            }
        }
    }
    encodeVal(obj) {
        const value = obj.value;
        const indexInLiteralsTable = this.literals.get(obj.value);
        if (indexInLiteralsTable !== undefined && (typeof value === 'number' || typeof value === 'string')) {
            this.u8(0xd6);
            this.ts(obj.id);
            this.ts(obj.writeId);
            this.vuint57(indexInLiteralsTable);
        }
        else {
            this.u8(0xd5);
            this.ts(obj.id);
            this.ts(obj.writeId);
            this.encodeAny(obj.value);
        }
    }
    encodeUtf8(str, byteLength) {
        this.ensureCapacity(byteLength);
        const strLength = str.length;
        const uint8 = this.uint8;
        let offset = this.offset;
        let pos = 0;
        while (pos < strLength) {
            let value = str.charCodeAt(pos++);
            if ((value & 0xffffff80) === 0) {
                uint8[offset++] = value;
                continue;
            }
            else if ((value & 0xfffff800) === 0) {
                uint8[offset++] = ((value >> 6) & 0x1f) | 0xc0;
            }
            else {
                if (value >= 0xd800 && value <= 0xdbff) {
                    if (pos < strLength) {
                        const extra = str.charCodeAt(pos);
                        if ((extra & 0xfc00) === 0xdc00) {
                            pos++;
                            value = ((value & 0x3ff) << 10) + (extra & 0x3ff) + 0x10000;
                        }
                    }
                }
                if ((value & 0xffff0000) === 0) {
                    uint8[offset++] = ((value >> 12) & 0x0f) | 0xe0;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
                else {
                    uint8[offset++] = ((value >> 18) & 0x07) | 0xf0;
                    uint8[offset++] = ((value >> 12) & 0x3f) | 0x80;
                    uint8[offset++] = ((value >> 6) & 0x3f) | 0x80;
                }
            }
            uint8[offset++] = (value & 0x3f) | 0x80;
        }
        this.offset = offset;
    }
}
exports.Encoder = Encoder;
