"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decoder = void 0;
const ArrayChunk_1 = require("../../types/rga-array/ArrayChunk");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const BinaryChunk_1 = require("../../types/rga-binary/BinaryChunk");
const BinaryType_1 = require("../../types/rga-binary/BinaryType");
const ClockDecoder_1 = require("../../../json-crdt-patch/codec/clock/ClockDecoder");
const ConstantType_1 = require("../../types/const/ConstantType");
const DocRootType_1 = require("../../types/lww-doc-root/DocRootType");
const constants_1 = require("../../constants");
const decode_1 = require("../../../util/base64/decode");
const clock_1 = require("../../../json-crdt-patch/clock");
const model_1 = require("../../model");
const ObjectChunk_1 = require("../../types/lww-object/ObjectChunk");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const constants_2 = require("../../../json-crdt-patch/constants");
const StringChunk_1 = require("../../types/rga-string/StringChunk");
const StringType_1 = require("../../types/rga-string/StringType");
const ValueType_1 = require("../../types/lww-value/ValueType");
class Decoder {
    decode(data) {
        const x = data[0];
        const isServerTime = typeof x === 'number';
        if (isServerTime) {
            this.time = x;
        }
        else {
            this.clockDecoder = ClockDecoder_1.ClockDecoder.fromArr(x);
        }
        const doc = isServerTime ? model_1.Model.withServerClock(x) : model_1.Model.withLogicalClock(this.clockDecoder.clock);
        const [rootId, index] = this.ts(data, 1);
        const rootNode = data[index] ? this.decodeNode(doc, data[index]) : null;
        doc.root = new DocRootType_1.DocRootType(doc, rootId, rootNode);
        return doc;
    }
    ts(arr, index) {
        const x = arr[index];
        if (typeof x === 'number') {
            if (x < 0) {
                const sessionIndex = -x;
                const timeDiff = arr[index + 1];
                return [this.clockDecoder.decodeId(sessionIndex, timeDiff), index + 2];
            }
            else {
                return [new clock_1.ServerTimestamp(this.time - x), index + 1];
            }
        }
        else {
            const time = x[0];
            switch (time) {
                case constants_2.ORIGIN.time:
                    return [constants_2.ORIGIN, index + 1];
                case constants_2.NULL_ID.time:
                    return [constants_2.NULL_ID, index + 1];
                case constants_2.TRUE_ID.time:
                    return [constants_2.TRUE_ID, index + 1];
                case constants_2.FALSE_ID.time:
                    return [constants_2.FALSE_ID, index + 1];
                case constants_2.UNDEFINED_ID.time:
                    return [constants_2.UNDEFINED_ID, index + 1];
                default:
                    return [new clock_1.LogicalTimestamp(0, time), index + 1];
            }
        }
    }
    decodeNode(doc, data) {
        switch (data) {
            case 1:
                return constants_1.NULL;
            case 2:
                return constants_1.FALSE;
            case 3:
                return constants_1.TRUE;
            case 4:
                return constants_1.UNDEFINED;
        }
        if (data instanceof Array) {
            switch (data[0]) {
                case 0:
                    return this.decodeObj(doc, data);
                case 1:
                    return this.decodeArr(doc, data);
                case 2:
                    return this.decodeStr(doc, data);
                case 3:
                    return this.decodeVal(doc, data);
                case 4:
                    return this.decodeConst(doc, data);
                case 5:
                    return this.decodeBin(doc, data);
            }
        }
        throw new Error('UNKNOWN_NODE');
    }
    decodeObj(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new ObjectType_1.ObjectType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const key = data[i];
            const [keyId, idx] = this.ts(data, i + 1);
            const node = this.decodeNode(doc, data[idx]);
            const chunk = new ObjectChunk_1.ObjectChunk(keyId, node);
            obj.putChunk(key, chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeArr(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new ArrayType_1.ArrayType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new ArrayChunk_1.ArrayChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                const nodes = content.map((c) => this.decodeNode(doc, c));
                chunk = new ArrayChunk_1.ArrayChunk(chunkId, nodes);
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeStr(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new StringType_1.StringType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new StringChunk_1.StringChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                chunk = new StringChunk_1.StringChunk(chunkId, content);
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeBin(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new BinaryType_1.BinaryType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new BinaryChunk_1.BinaryChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                chunk = new BinaryChunk_1.BinaryChunk(chunkId, (0, decode_1.fromBase64)(content));
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeVal(doc, data) {
        const [id, index1] = this.ts(data, 1);
        const [writeId, index2] = this.ts(data, index1);
        const value = data[index2];
        const obj = new ValueType_1.ValueType(id, writeId, value);
        doc.nodes.index(obj);
        return obj;
    }
    decodeConst(doc, data) {
        const value = data[1];
        const obj = new ConstantType_1.ConstantType(constants_2.ORIGIN, value);
        return obj;
    }
}
exports.Decoder = Decoder;
