"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryType = void 0;
const BinaryChunk_1 = require("./BinaryChunk");
const BinaryOriginChunk_1 = require("./BinaryOriginChunk");
class BinaryType {
    constructor(doc, id) {
        this.doc = doc;
        this.id = id;
        this.start = this.end = new BinaryOriginChunk_1.BinaryOriginChunk(id);
    }
    onInsert(op) {
        delete this._toJson;
        let curr = this.end;
        while (curr) {
            if (curr.id.overlap(curr.span(), op.id, op.span()))
                return;
            if (curr.id.inSpan(curr.span(), op.after, 1))
                break;
            curr = curr.left;
        }
        if (!curr)
            return;
        const isOriginChunk = curr instanceof BinaryOriginChunk_1.BinaryOriginChunk;
        if (!curr.deleted && !isOriginChunk) {
            const doesAfterMatch = curr.id.getSessionId() === op.after.getSessionId() && curr.id.time + curr.span() - 1 === op.after.time;
            const isIdSameSession = curr.id.getSessionId() === op.id.getSessionId();
            const isIdIncreasingWithoutAGap = curr.id.time + curr.span() === op.id.time;
            if (doesAfterMatch && isIdSameSession && isIdIncreasingWithoutAGap) {
                curr.merge(op.data);
                return;
            }
        }
        const chunk = new BinaryChunk_1.BinaryChunk(op.id, op.data);
        const targetsLastElementInChunk = op.after.time === curr.id.time + curr.span() - 1;
        if (targetsLastElementInChunk) {
            while (curr.right && curr.right.id.compare(op.id) > 0)
                curr = curr.right;
            this.insertChunk(chunk, curr);
            return;
        }
        this.splitChunk(curr, op.after.time);
        this.insertChunk(chunk, curr);
    }
    onDelete(op) {
        delete this._toJson;
        const { after } = op;
        const length = after.span;
        let chunk = this.end;
        const chunks = [];
        while (chunk) {
            if (chunk.id.overlap(chunk.span(), after, length))
                chunks.push(chunk);
            if (chunk.id.inSpan(chunk.span(), after, 1))
                break;
            chunk = chunk.left;
        }
        for (const c of chunks)
            this.deleteInChunk(c, after.time, after.time + length - 1);
    }
    insertChunk(chunk, after) {
        chunk.left = after;
        chunk.right = after.right;
        if (after.right)
            after.right.left = chunk;
        else
            this.end = chunk;
        after.right = chunk;
    }
    splitChunk(chunk, time) {
        const newChunk = chunk.split(time);
        this.insertChunk(newChunk, chunk);
        return newChunk;
    }
    deleteInChunk(chunk, t1, t2) {
        const c1 = chunk.id.time;
        const c2 = c1 + chunk.span() - 1;
        if (t1 <= c1) {
            if (t2 >= c2)
                chunk.delete();
            else {
                this.splitChunk(chunk, t2);
                chunk.delete();
            }
        }
        else {
            if (t2 >= c2)
                this.splitChunk(chunk, t1 - 1).delete();
            else {
                this.splitChunk(chunk, t2);
                this.splitChunk(chunk, t1 - 1).delete();
            }
        }
    }
    findId(index) {
        let chunk = this.start;
        let cnt = 0;
        const next = index + 1;
        while (chunk) {
            if (chunk.buf) {
                cnt += chunk.buf.length;
                if (cnt >= next)
                    return chunk.id.tick(chunk.buf.length - (cnt - index));
            }
            chunk = chunk.right;
        }
        throw new Error('OUT_OF_BOUNDS');
    }
    findIdSpan(index, length) {
        let chunk = this.start;
        let cnt = 0;
        const next = index + 1;
        while (chunk) {
            if (chunk.buf) {
                cnt += chunk.buf.length;
                if (cnt >= next) {
                    const remaining = cnt - index;
                    if (remaining >= length)
                        return [chunk.id.interval(chunk.span() - remaining, length)];
                    length -= remaining;
                    const result = [chunk.id.interval(chunk.span() - remaining, remaining)];
                    while (chunk.right) {
                        chunk = chunk.right;
                        if (chunk.deleted)
                            continue;
                        const span = chunk.span();
                        if (span >= length) {
                            result.push(chunk.id.interval(0, length));
                            break;
                        }
                        result.push(chunk.id.interval(0, span));
                        length -= span;
                    }
                    return result;
                }
            }
            chunk = chunk.right;
        }
        throw new Error('OUT_OF_BOUNDS');
    }
    append(chunk) {
        const last = this.end;
        last.right = chunk;
        chunk.left = last;
        this.end = chunk;
    }
    toJson() {
        if (this._toJson)
            return this._toJson;
        let curr = this.start;
        const buffers = [];
        let length = 0;
        while ((curr = curr.right)) {
            const buf = curr.buf;
            if (buf) {
                buffers.push(buf);
                length += buf.length;
            }
        }
        const res = new Uint8Array(length);
        let offset = 0;
        for (let i = 0; i < buffers.length; i++) {
            const buf = buffers[i];
            res.set(buf, offset);
            offset += buf.length;
        }
        return this._toJson = res;
    }
    clone(doc) {
        const copy = new BinaryType(this.doc, this.id);
        let i = this.start;
        let j = copy.start;
        while (i.right) {
            const cloned = i.right.clone();
            j.right = cloned;
            cloned.left = j;
            j = cloned;
            i = i.right;
        }
        copy.end = j;
        doc.nodes.index(copy);
        return copy;
    }
    *children() { }
    *chunks() {
        let curr = this.start;
        while ((curr = curr.right))
            yield curr;
    }
    size() {
        let curr = this.start;
        let size = 0;
        while ((curr = curr.right))
            size++;
        return size;
    }
    length() {
        let curr = this.start;
        let size = 0;
        while ((curr = curr.right))
            if (curr.buf)
                size += curr.buf.length;
        return size;
    }
    toString(tab = '') {
        let str = `${tab}${this.constructor.name}(${this.id.toDisplayString()})`;
        let curr = this.start;
        while (curr) {
            str += `\n${curr.toString(tab + '  ')}`;
            curr = curr.right;
        }
        return str;
    }
}
exports.BinaryType = BinaryType;
