"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonSerializerCodegen = void 0;
const codegen_1 = require("../util/codegen");
const JsExpression_1 = require("../util/codegen/util/JsExpression");
const asString_1 = require("../util/asString");
const normalizeAccessor_1 = require("../util/codegen/util/normalizeAccessor");
const encode_1 = require("../util/base64/encode");
class WriteTextStep {
    constructor(str) {
        this.str = str;
    }
}
class JsonSerializerCodegen {
    constructor(opts) {
        this.base64Linked = false;
        this.options = {
            ref: (id) => undefined,
            ...opts,
        };
        const typeNamePart = this.options.type.id && /^[a-z][a-z0-9_]*$/i.test(this.options.type.id)
            ? this.options.type.id[0].toUpperCase() + this.options.type.id.substr(1)
            : '';
        this.codegen = new codegen_1.Codegen({
            name: 'toJson' + typeNamePart,
            prologue: `var s = '';`,
            epilogue: `return s;`,
            processSteps: (steps) => {
                const stepsJoined = [];
                for (let i = 0; i < steps.length; i++) {
                    const step = steps[i];
                    if (step instanceof codegen_1.CodegenStepExecJs)
                        stepsJoined.push(step);
                    else if (step instanceof WriteTextStep) {
                        const last = stepsJoined[stepsJoined.length - 1];
                        if (last instanceof WriteTextStep)
                            last.str += step.str;
                        else
                            stepsJoined.push(step);
                    }
                }
                const execSteps = [];
                for (const step of stepsJoined) {
                    if (step instanceof codegen_1.CodegenStepExecJs) {
                        execSteps.push(step);
                    }
                    else if (step instanceof WriteTextStep) {
                        const js = `s += ${JSON.stringify(step.str)};`;
                        execSteps.push(new codegen_1.CodegenStepExecJs(js));
                    }
                }
                return execSteps;
            },
        });
        this.codegen.linkDependency(asString_1.asString, 'asString');
        this.codegen.linkDependency(JSON.stringify, 'stringify');
    }
    linkBase64() {
        if (this.base64Linked)
            return;
        this.codegen.linkDependency(encode_1.toBase64, 'toBase64');
    }
    js(js) {
        this.codegen.js(js);
    }
    writeText(str) {
        this.codegen.step(new WriteTextStep(str));
    }
    onString(str, value) {
        if (str.const !== undefined) {
            this.writeText(JSON.stringify(str.const));
            return;
        }
        if (str.noJsonEscape) {
            this.js(`s += ${value.use()};`);
            return;
        }
        this.js(`s += asString(${value.use()});`);
    }
    onNumber(num, value) {
        if (num.const !== undefined) {
            this.writeText(JSON.stringify(num.const));
            return;
        }
        this.js(`s += ${value.use()};`);
    }
    onBoolean(bool, value) {
        if (bool.const !== undefined) {
            this.writeText(JSON.stringify(bool.const));
            return;
        }
        this.js(`s += ${value.use()} ? 'true' : 'false';`);
    }
    onNull() {
        this.writeText('null');
    }
    onArray(arr, value) {
        if (arr.const !== undefined) {
            this.writeText(JSON.stringify(arr.const));
            return;
        }
        this.writeText('[');
        const r = this.codegen.getRegister();
        const rl = this.codegen.getRegister();
        const ri = this.codegen.getRegister();
        const rItem = this.codegen.getRegister();
        this.js(`var ${r} = ${value.use()}, ${rl} = ${r}.length, ${ri} = 0, ${rItem};`);
        this.js(`for(; ${ri} < ${rl}; ${ri}++) ` + '{');
        this.js(`${rItem} = ${r}[${ri}];`);
        this.js(`if (${ri}) s += ',';`);
        this.onType(arr.type, new JsExpression_1.JsExpression(() => `${rItem}`));
        this.js(`}`);
        this.writeText(']');
    }
    onObject(obj, value) {
        if (obj.unknownFields) {
            this.js(`s += stringify(${value.use()});`);
            return;
        }
        const requiredFields = obj.fields.filter((field) => !field.isOptional);
        const optionalFields = obj.fields.filter((field) => field.isOptional);
        this.writeText('{');
        for (let i = 0; i < requiredFields.length; i++) {
            const field = requiredFields[i];
            if (i)
                this.writeText(',');
            this.writeText(JSON.stringify(field.key) + ':');
            const accessor = (0, normalizeAccessor_1.normalizeAccessor)(field.key);
            this.onType(field.type, new JsExpression_1.JsExpression(() => `${value.use()}${accessor}`));
        }
        const rHasFields = this.codegen.getRegister();
        if (!requiredFields.length) {
            this.js(`var ${rHasFields} = false;`);
        }
        for (let i = 0; i < optionalFields.length; i++) {
            const field = optionalFields[i];
            const accessor = (0, normalizeAccessor_1.normalizeAccessor)(field.key);
            const rValue = this.codegen.getRegister();
            this.js(`var ${rValue} = ${value.use()}${accessor};`);
            this.js(`if (${rValue} !== undefined) {`);
            if (requiredFields.length) {
                this.writeText(',');
            }
            else {
                this.js(`if (${rHasFields}) s += ',';`);
                this.js(`${rHasFields} = true;`);
            }
            this.writeText(JSON.stringify(field.key) + ':');
            this.onType(field.type, new JsExpression_1.JsExpression(() => `${rValue}`));
            this.js(`}`);
        }
        this.writeText('}');
    }
    onBinary(value) {
        this.linkBase64();
        this.writeText('"data:application/octet-stream;base64,');
        this.js(`s += toBase64(${value.use()});`);
        this.writeText('"');
    }
    onRef(ref, value) {
        const type = this.options.ref(ref.ref);
        switch (typeof type) {
            case 'function': {
                const d = this.codegen.linkDependency(type);
                this.js(`s += ${d}(${value.use()});`);
                break;
            }
            case 'object': {
                this.onType(type, value);
                break;
            }
            default:
                throw new Error(`Unknown [ref = ${ref.ref}].`);
        }
    }
    onAny(value) {
        this.js(`s += stringify(${value.use()});`);
    }
    onType(type, value) {
        switch (type.__t) {
            case 'str': {
                this.onString(type, value);
                break;
            }
            case 'num': {
                this.onNumber(type, value);
                break;
            }
            case 'bool': {
                this.onBoolean(type, value);
                break;
            }
            case 'nil': {
                this.onNull();
                break;
            }
            case 'arr': {
                this.onArray(type, value);
                break;
            }
            case 'obj': {
                this.onObject(type, value);
                break;
            }
            case 'bin': {
                this.onBinary(value);
                break;
            }
            case 'ref': {
                this.onRef(type, value);
                break;
            }
            default: {
                this.onAny(value);
                break;
            }
        }
    }
    run() {
        const r = this.codegen.getRegister();
        const value = new JsExpression_1.JsExpression(() => r);
        this.onType(this.options.type, value);
        return this;
    }
    generate() {
        return this.codegen.generate();
    }
    compile() {
        return this.codegen.compile();
    }
}
exports.JsonSerializerCodegen = JsonSerializerCodegen;
