"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTypeSystem = void 0;
const json_type_validator_1 = require("../json-type-validator");
const dynamicFunction_1 = require("../util/codegen/dynamicFunction");
const json_type_serializer_1 = require("../json-type-serializer");
const util_1 = require("../json-pack/util");
const json_type_schema_1 = require("../json-type-schema");
const toTypeScriptText_1 = require("../json-type-typescript/toTypeScriptText");
const export_1 = require("../json-type-typescript/export");
const literalTypes = ['bin', 'bool', 'nil', 'num', 'str'];
class JsonTypeSystem {
    constructor(options) {
        this.options = options;
        this.ref = (ref) => {
            const type = this.options.types[ref];
            if (!type)
                throw new Error(`Type [ref = ${ref}] not found.`);
            return type;
        };
        this.fastValidatorCache = {};
        this.fastValidatorUsage = {};
        this.fullValidatorCache = {};
        this.jsonSerializerCache = {};
        this.jsonSerializerUsage = {};
        this.msgPackPartialEncoderCache = {};
        this.msgPackPartialEncoderUsage = {};
        this.msgPackEncoderCache = {};
    }
    hasType(ref) {
        return Object.prototype.hasOwnProperty.call(this.options.types, ref);
    }
    getFastValidator(ref) {
        const type = this.ref(ref);
        const cachedFastValidator = this.fastValidatorCache[ref];
        if (cachedFastValidator) {
            this.fastValidatorUsage[ref] = (this.fastValidatorUsage[ref] || 0) + 1;
            return cachedFastValidator;
        }
        const [validator, setValidator] = (0, dynamicFunction_1.dynamicFunction)(() => {
            throw new Error(`Type [ref = ${ref}] not implemented.`);
        });
        this.fastValidatorCache[ref] = validator;
        const realValidator = (0, json_type_validator_1.createBoolValidator)(type, {
            customValidators: this.options.customValidators,
            ref: (id) => this.getFastValidator(id),
            skipObjectExtraFieldsCheck: true,
            unsafeMode: true,
        });
        if (this.fastValidatorUsage[ref])
            setValidator(realValidator);
        this.fastValidatorCache[ref] = realValidator;
        return realValidator;
    }
    getFullValidator(ref) {
        const type = this.ref(ref);
        const cachedFullValidator = this.fullValidatorCache[ref];
        if (cachedFullValidator)
            return cachedFullValidator;
        const [validator, setValidator] = (0, dynamicFunction_1.dynamicFunction)(() => {
            throw new Error(`Type [ref = ${ref}] not implemented.`);
        });
        this.fullValidatorCache[ref] = validator;
        setValidator((0, json_type_validator_1.createObjValidator)(type, {
            customValidators: this.options.customValidators,
            ref: (id) => this.getFullValidator(id),
        }));
        return validator;
    }
    getJsonSerializer(ref) {
        const type = this.ref(ref);
        const jsonSerializerCached = this.jsonSerializerCache[ref];
        if (jsonSerializerCached) {
            this.jsonSerializerUsage[ref] = (this.jsonSerializerUsage[ref] || 0) + 1;
            return jsonSerializerCached;
        }
        const codegen = new json_type_serializer_1.JsonSerializerCodegen({
            type,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getJsonSerializer(id);
            },
        });
        const [serializer, setSerializer] = (0, dynamicFunction_1.dynamicFunction)(() => {
            throw new Error(`Type [ref = ${ref}] JSON serializer not implemented.`);
        });
        this.jsonSerializerCache[ref] = serializer;
        const realSerializer = codegen.run().compile();
        if (this.jsonSerializerUsage[ref])
            setSerializer(realSerializer);
        this.jsonSerializerCache[ref] = realSerializer;
        return realSerializer;
    }
    getMsgPackPartialEncoder(ref) {
        const type = this.ref(ref);
        const msgPackPartialEncoderCached = this.msgPackPartialEncoderCache[ref];
        if (msgPackPartialEncoderCached) {
            this.msgPackPartialEncoderUsage[ref] = (this.msgPackPartialEncoderUsage[ref] || 0) + 1;
            return msgPackPartialEncoderCached;
        }
        const codegen = new json_type_serializer_1.MsgPackSerializerCodegen({
            type,
            encoder: util_1.encoderFull,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getMsgPackPartialEncoder(id);
            },
        });
        const [partialEncoder, setPartialEncoder] = (0, dynamicFunction_1.dynamicFunction)(() => {
            throw new Error(`Type [ref = ${ref}] MessagePack partial encoder not implemented.`);
        });
        this.msgPackPartialEncoderCache[ref] = partialEncoder;
        const realPartialEncoder = codegen.run().compilePartial();
        if (this.msgPackPartialEncoderUsage[ref])
            setPartialEncoder(realPartialEncoder);
        this.msgPackPartialEncoderCache[ref] = realPartialEncoder;
        return realPartialEncoder;
    }
    getMsgPackEncoder(ref) {
        const type = this.ref(ref);
        const msgPackSerializerCached = this.msgPackEncoderCache[ref];
        if (msgPackSerializerCached)
            return msgPackSerializerCached;
        const codegen = new json_type_serializer_1.MsgPackSerializerCodegen({
            type,
            encoder: util_1.encoderFull,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getMsgPackPartialEncoder(id);
            },
        });
        return (this.msgPackEncoderCache[ref] = codegen.run().compile());
    }
    toJsonSchema(ref, withDefs) {
        const type = this.ref(ref);
        const defs = withDefs ? {} : undefined;
        let schema = (0, json_type_schema_1.toJsonSchema)(type, {
            ref: this.ref,
            defs,
        });
        if (defs) {
            schema = {
                $id: ref,
                ...schema,
                $defs: defs,
            };
        }
        return schema;
    }
    toTypeScriptText(ref) {
        const type = this.ref(ref);
        return (0, toTypeScriptText_1.toTypeScriptText)(type, '');
    }
    toTsAst(ref) {
        const statements = [];
        if (ref) {
            (0, export_1.exportDeclaration)(ref, {
                ref: this.ref,
                statements,
            });
        }
        else {
            const types = this.options.types;
            for (const ref of Object.keys(types)) {
                (0, export_1.exportDeclaration)(ref, {
                    ref: this.ref,
                    statements,
                });
            }
        }
        return statements;
    }
}
exports.JsonTypeSystem = JsonTypeSystem;
