"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcApiCaller = void 0;
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const constants_1 = require("./constants");
const BufferSubject_1 = require("../../../util/BufferSubject");
const error_1 = require("./error");
const methods_1 = require("./methods");
class RpcApiCaller {
    constructor({ types, api, error, preCallBufferSize = 10, maxActiveCalls = 50 }) {
        this._calls = 0;
        this.api = {};
        for (const key in api)
            this.api[key] = (0, methods_1.wrapMethod)(api[key], types);
        this.error = error || new error_1.ErrorLikeErrorFormatter();
        this.preCallBufferSize = preCallBufferSize;
        this.maxActiveCalls = maxActiveCalls;
    }
    get activeCalls() {
        return this._calls;
    }
    exists(name) {
        return this.api.hasOwnProperty(name);
    }
    get(name) {
        if (!this.exists(name))
            throw new error_1.RpcError(constants_1.RpcServerError.NoMethodSpecified);
        return this.api[name];
    }
    async call(name, request, ctx) {
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (method.validate)
                method.validate(request);
            try {
                this._calls++;
                if (method.onPreCall)
                    await method.onPreCall(ctx, request);
                return await method.call(ctx, request);
            }
            finally {
                this._calls--;
            }
        }
        catch (error) {
            throw this.error.format(error);
        }
    }
    async callJson(name, request, ctx) {
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (method.validate)
                method.validate(request);
            try {
                this._calls++;
                if (method.onPreCall)
                    await method.onPreCall(ctx, request);
                return await method.callJson(ctx, request);
            }
            finally {
                this._calls--;
            }
        }
        catch (error) {
            throw this.error.format(error);
        }
    }
    async callMsgPack(name, request, ctx) {
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (method.validate)
                method.validate(request);
            try {
                this._calls++;
                if (method.onPreCall)
                    await method.onPreCall(ctx, request);
                return await method.callMsgPack(ctx, request);
            }
            finally {
                this._calls--;
            }
        }
        catch (error) {
            throw this.error.format(error);
        }
    }
    createCall(name, ctx, callMethod, callMethod$) {
        const req$ = new rxjs_1.Subject();
        const reqUnsubscribe$ = new rxjs_1.Subject();
        try {
            if (this._calls >= this.maxActiveCalls)
                throw new error_1.RpcError(constants_1.RpcServerError.TooManyActiveCalls);
            const method = this.get(name);
            if (!method.isStreaming) {
                const response$ = (0, rxjs_1.from)((async () => {
                    const request = await (0, rxjs_1.firstValueFrom)(req$.pipe((0, operators_1.first)()));
                    const response = await callMethod(name, request, ctx);
                    return response;
                })());
                const res$ = new rxjs_1.Subject();
                response$.subscribe(res$);
                const $resWithErrorsFormatted = res$.pipe((0, operators_1.catchError)((error) => {
                    throw this.error.format(error);
                }));
                return { req$, reqUnsubscribe$, res$: $resWithErrorsFormatted };
            }
            const methodStreaming = method;
            const requestValidated$ = req$.pipe((0, operators_1.map)((request) => {
                if (methodStreaming.validate)
                    methodStreaming.validate(request);
                return request;
            }));
            const bufferSize = methodStreaming.preCallBufferSize || this.preCallBufferSize;
            const requestBuffered$ = new BufferSubject_1.BufferSubject(bufferSize);
            const error$ = new rxjs_1.Subject();
            requestBuffered$.subscribe({
                error: (error) => {
                    error$.error(error);
                },
            });
            requestValidated$.subscribe(requestBuffered$);
            const result$ = requestBuffered$.pipe((0, operators_1.take)(1), (0, operators_1.switchMap)((request) => {
                return methodStreaming.onPreCall ? (0, rxjs_1.from)(methodStreaming.onPreCall(ctx, request)) : (0, rxjs_1.from)([0]);
            }), (0, operators_1.switchMap)(() => {
                Promise.resolve().then(() => {
                    requestBuffered$.flush();
                });
                return callMethod$(methodStreaming, ctx, requestBuffered$).pipe((0, operators_1.finalize)(() => {
                    error$.complete();
                }));
            }), (0, operators_1.share)(), (0, operators_1.mergeWith)(error$));
            const resultWithActiveCallTracking$ = (0, rxjs_1.of)(null).pipe((0, operators_1.tap)(() => {
                this._calls++;
            }), (0, operators_1.switchMap)(() => result$), (0, operators_1.finalize)(() => {
                error$.complete();
                this._calls--;
            }));
            const $resWithErrorsFormatted = resultWithActiveCallTracking$.pipe((0, operators_1.catchError)((error) => {
                throw this.error.format(error);
            }));
            return { req$, reqUnsubscribe$, res$: $resWithErrorsFormatted };
        }
        catch (error) {
            const errorFormatted = this.error.format(error);
            req$.error(errorFormatted);
            const res$ = new rxjs_1.Subject();
            res$.error(errorFormatted);
            return { req$, reqUnsubscribe$, res$ };
        }
    }
    call$(name, request$, ctx) {
        const call = this.createCall(name, ctx, (name, req, ctx) => this.call(name, req, ctx), (method, ctx, req$) => method.call$(ctx, req$));
        request$.subscribe(call.req$);
        return call.res$;
    }
    callJson$(name, request$, ctx) {
        const call = this.createCall(name, ctx, (name, req, ctx) => this.callJson(name, req, ctx), (method, ctx, req$) => method.callJson$(ctx, req$));
        request$.subscribe(call.req$);
        return call.res$;
    }
    callMsgPack$(name, request$, ctx) {
        const call = this.createCall(name, ctx, (name, req, ctx) => this.callMsgPack(name, req, ctx), (method, ctx, req$) => method.callMsgPack$(ctx, req$));
        request$.subscribe(call.req$);
        return call.res$;
    }
}
exports.RpcApiCaller = RpcApiCaller;
