"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryRxClient = void 0;
const messages_1 = require("./messages");
const rxjs_1 = require("rxjs");
const TimedQueue_1 = require("../json-rx/TimedQueue");
const codec_1 = require("./codec");
class BinaryRxClient {
    constructor({ send, bufferSize = 100, bufferTime = 10 }) {
        this.id = 1;
        this.encoder = new codec_1.Encoder();
        this.observers = new Map();
        this.buffer = new TimedQueue_1.TimedQueue();
        this.buffer.itemLimit = bufferSize;
        this.buffer.timeLimit = bufferTime;
        this.buffer.onFlush = (messages) => {
            send(this.encoder.encode(messages));
        };
    }
    sendSubscribe(id, method, data) {
        const message = new messages_1.SubscribeMessage(id, method, data);
        this.buffer.push(message);
    }
    sendUnsubscribe(id) {
        this.buffer.push(new messages_1.UnsubscribeMessage(id));
    }
    onData(message) {
        const { id, data } = message;
        const observer = this.observers.get(id);
        if (!observer)
            return;
        observer.observer.next(data);
    }
    onComplete(message) {
        const { id, data } = message;
        const observer = this.observers.get(id);
        if (!observer)
            return;
        observer.completed = true;
        if (data && data.byteLength)
            observer.observer.next(data);
        observer.observer.complete();
    }
    onError(message) {
        const { id, data } = message;
        const observer = this.observers.get(id);
        if (!observer)
            return;
        observer.completed = true;
        observer.observer.error(data);
    }
    onArray(arr) {
        const messages = (0, codec_1.decodeFullMessages)(arr, 0);
        this.onMessages(messages);
    }
    onMessages(messages) {
        const length = messages.length;
        for (let i = 0; i < length; i++) {
            const message = messages[i];
            if (message instanceof messages_1.CompleteMessage)
                this.onComplete(message);
            else if (message instanceof messages_1.DataMessage)
                this.onData(message);
            else if (message instanceof messages_1.ErrorMessage)
                this.onError(message);
        }
    }
    call(method, data) {
        const id = this.id++;
        if (this.id >= 0xffff)
            this.id = 1;
        if (this.observers.has(id))
            return this.call(method, data);
        const observable = new rxjs_1.Observable((observer) => {
            const entry = { observer };
            this.observers.set(id, entry);
            return () => {
                this.observers.delete(id);
                if (!entry.completed)
                    this.sendUnsubscribe(id);
            };
        });
        this.sendSubscribe(id, method, data);
        return observable;
    }
    notify(method, data) {
        const message = new messages_1.NotificationMessage(method, data);
        this.buffer.push(message);
    }
    stop() {
        this.buffer.onFlush = (message) => { };
        for (const { observer } of this.observers.values())
            observer.complete();
    }
}
exports.BinaryRxClient = BinaryRxClient;
