"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.stringify = exports.parse = void 0;
const json_pack_1 = require("../json-pack");
const base64_1 = require("../util/base64");
const isUint8Array_1 = require("../util/isUint8Array");
const constants_1 = require("./constants");
const binUriStartLength = constants_1.binUriStart.length;
const msgPackUriStartLength = constants_1.msgPackUriStart.length;
const msgPackExtStartLength = constants_1.msgPackExtStart.length;
const minDataUri = Math.min(binUriStartLength, msgPackUriStartLength);
const parseExtDataUri = (uri) => {
    uri = uri.substring(msgPackExtStartLength);
    const commaIndex = uri.indexOf(',');
    if (commaIndex === -1)
        throw new Error('INVALID_EXT_DATA_URI');
    const typeString = uri.substring(0, commaIndex);
    const buf = (0, base64_1.fromBase64)(uri.substring(commaIndex + 1));
    return new json_pack_1.JsonPackExtension(Number(typeString), buf);
};
const unwrapBinary = (value) => {
    if (!value)
        return value;
    if (value instanceof Array) {
        const len = value.length;
        for (let i = 0; i < len; i++) {
            const item = value[i];
            switch (typeof item) {
                case 'object': {
                    unwrapBinary(item);
                    continue;
                }
                case 'string': {
                    if (item.length < minDataUri)
                        continue;
                    if (item.substring(0, binUriStartLength) === constants_1.binUriStart)
                        value[i] = (0, base64_1.fromBase64)(item.substring(binUriStartLength));
                    else if (item.substring(0, msgPackUriStartLength) === constants_1.msgPackUriStart)
                        value[i] = new json_pack_1.JsonPackValue((0, base64_1.fromBase64)(item.substring(msgPackUriStartLength)));
                    else if (item.substring(0, msgPackExtStartLength) === constants_1.msgPackExtStart)
                        value[i] = parseExtDataUri(item);
                }
            }
        }
        return value;
    }
    if (typeof value === 'object') {
        for (const key in value) {
            const item = value[key];
            switch (typeof item) {
                case 'object': {
                    unwrapBinary(item);
                    continue;
                }
                case 'string': {
                    if (item.length < minDataUri)
                        continue;
                    if (item.substring(0, binUriStartLength) === constants_1.binUriStart) {
                        const buf = (0, base64_1.fromBase64)(item.substring(binUriStartLength));
                        value[key] = buf;
                    }
                    else if (item.substring(0, msgPackUriStartLength) === constants_1.msgPackUriStart) {
                        value[key] = new json_pack_1.JsonPackValue((0, base64_1.fromBase64)(item.substring(msgPackUriStartLength)));
                    }
                    else if (item.substring(0, msgPackExtStartLength) === constants_1.msgPackExtStart)
                        value[key] = parseExtDataUri(item);
                }
            }
        }
        return value;
    }
    if (typeof value === 'string') {
        if (value.length < minDataUri)
            return value;
        if (value.substring(0, binUriStartLength) === constants_1.binUriStart)
            return (0, base64_1.fromBase64)(value.substring(binUriStartLength));
        if (value.substring(0, msgPackUriStartLength) === constants_1.msgPackUriStart)
            return new json_pack_1.JsonPackValue((0, base64_1.fromBase64)(value.substring(msgPackUriStartLength)));
        if (value.substring(0, msgPackExtStartLength) === constants_1.msgPackExtStart)
            return parseExtDataUri(value);
        else
            return value;
    }
    return value;
};
const parse = (json) => {
    const parsed = JSON.parse(json);
    return unwrapBinary(parsed);
};
exports.parse = parse;
const wrapBinary = (value) => {
    if (!value)
        return value;
    if ((0, isUint8Array_1.isUint8Array)(value))
        return constants_1.binUriStart + (0, base64_1.toBase64)(value);
    if (value instanceof Array) {
        const out = [];
        const len = value.length;
        for (let i = 0; i < len; i++) {
            const item = value[i];
            out.push(!item || typeof item !== 'object' ? item : wrapBinary(item));
        }
        return out;
    }
    if (value instanceof json_pack_1.JsonPackValue)
        return constants_1.msgPackUriStart + (0, base64_1.toBase64)(value.buf);
    if (value instanceof json_pack_1.JsonPackExtension)
        return constants_1.msgPackExtStart + value.type + ',' + (0, base64_1.toBase64)(value.buf);
    if (typeof value === 'object') {
        const out = {};
        for (const key in value) {
            const item = value[key];
            out[key] = !item || typeof item !== 'object' ? item : wrapBinary(item);
        }
        return out;
    }
    return value;
};
const stringify = (value, replacer, space) => {
    const wrapped = wrapBinary(value);
    return JSON.stringify(wrapped, replacer, space);
};
exports.stringify = stringify;
