"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Encoder = void 0;
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const BinaryType_1 = require("../../types/rga-binary/BinaryType");
const ClockEncoder_1 = require("../../../json-crdt-patch/codec/clock/ClockEncoder");
const ConstantType_1 = require("../../types/const/ConstantType");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const StringType_1 = require("../../types/rga-string/StringType");
const encode_1 = require("../../../util/base64/encode");
const ValueType_1 = require("../../types/lww-value/ValueType");
class Encoder {
    encode(model) {
        const isServerTime = model.clock.getSessionId() === 1;
        const clock = model.clock;
        const snapshot = isServerTime ? [clock.time] : [null];
        if (isServerTime) {
            this.time = clock.time;
        }
        else {
            model.advanceClocks();
            this.clock = new ClockEncoder_1.ClockEncoder(model.clock);
        }
        this.encodeRoot(snapshot, model.root);
        if (!isServerTime)
            snapshot[0] = this.clock.toJson();
        return snapshot;
    }
    ts(arr, ts) {
        switch (ts.getSessionId()) {
            case 0: {
                arr.push([ts.time]);
                break;
            }
            case 1: {
                arr.push(this.time - ts.time);
                break;
            }
            default: {
                const relativeId = this.clock.append(ts);
                arr.push(-relativeId.sessionIndex, relativeId.timeDiff);
            }
        }
    }
    encodeRoot(arr, root) {
        this.ts(arr, root.id);
        if (!root.node)
            arr.push(0);
        else
            this.encodeNode(arr, root.node);
    }
    encodeNode(arr, node) {
        if (node instanceof ObjectType_1.ObjectType)
            return this.encodeObj(arr, node);
        else if (node instanceof ArrayType_1.ArrayType)
            return this.encodeArr(arr, node);
        else if (node instanceof StringType_1.StringType)
            return this.encodeStr(arr, node);
        else if (node instanceof ValueType_1.ValueType)
            return this.encodeVal(arr, node);
        else if (node instanceof ConstantType_1.ConstantType)
            return this.encodeConst(arr, node);
        else if (node instanceof BinaryType_1.BinaryType)
            return this.encodeBin(arr, node);
        throw new Error('UNKNOWN_NODE');
    }
    encodeObj(arr, obj) {
        const res = [0];
        arr.push(res);
        this.ts(res, obj.id);
        for (const [key, entry] of obj.latest.entries()) {
            res.push(key);
            this.ts(res, entry.id);
            this.encodeNode(res, entry.node);
        }
    }
    encodeArr(arr, obj) {
        const res = [1];
        arr.push(res);
        this.ts(res, obj.id);
        for (const chunk of obj.chunks())
            this.encodeArrChunk(res, chunk);
    }
    encodeArrChunk(arr, chunk) {
        this.ts(arr, chunk.id);
        if (chunk.deleted)
            arr.push(chunk.deleted);
        else {
            const nodes = [];
            for (const n of chunk.nodes)
                this.encodeNode(nodes, n);
            arr.push(nodes);
        }
    }
    encodeStr(arr, obj) {
        const res = [2];
        arr.push(res);
        this.ts(res, obj.id);
        for (const chunk of obj.chunks())
            this.encodeStrChunk(res, chunk);
    }
    encodeStrChunk(arr, chunk) {
        this.ts(arr, chunk.id);
        if (chunk.deleted)
            arr.push(chunk.deleted);
        else
            arr.push(chunk.str);
    }
    encodeBin(arr, obj) {
        const res = [5];
        arr.push(res);
        this.ts(res, obj.id);
        for (const chunk of obj.chunks())
            this.encodeBinChunk(res, chunk);
    }
    encodeBinChunk(arr, chunk) {
        this.ts(arr, chunk.id);
        if (chunk.deleted)
            arr.push(chunk.deleted);
        else
            arr.push((0, encode_1.toBase64)(chunk.buf));
    }
    encodeVal(arr, obj) {
        const res = [3];
        arr.push(res);
        this.ts(res, obj.id);
        this.ts(res, obj.writeId);
        res.push(obj.value);
    }
    encodeConst(arr, obj) {
        switch (obj.value) {
            case null: {
                arr.push(1);
                return;
            }
            case false: {
                arr.push(2);
                return;
            }
            case true: {
                arr.push(3);
                return;
            }
            case undefined: {
                arr.push(4);
                return;
            }
        }
        const res = [4];
        arr.push(res);
        res.push(obj.value);
    }
}
exports.Encoder = Encoder;
