"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decoder = void 0;
const ArrayChunk_1 = require("../../types/rga-array/ArrayChunk");
const ArrayType_1 = require("../../types/rga-array/ArrayType");
const BinaryChunk_1 = require("../../types/rga-binary/BinaryChunk");
const BinaryType_1 = require("../../types/rga-binary/BinaryType");
const ConstantType_1 = require("../../types/const/ConstantType");
const DocRootType_1 = require("../../types/lww-doc-root/DocRootType");
const constants_1 = require("../../constants");
const decode_1 = require("../../../util/base64/decode");
const clock_1 = require("../../../json-crdt-patch/clock");
const model_1 = require("../../model");
const ObjectChunk_1 = require("../../types/lww-object/ObjectChunk");
const ObjectType_1 = require("../../types/lww-object/ObjectType");
const constants_2 = require("../../../json-crdt-patch/constants");
const StringChunk_1 = require("../../types/rga-string/StringChunk");
const StringType_1 = require("../../types/rga-string/StringType");
const ValueType_1 = require("../../types/lww-value/ValueType");
class Decoder {
    decode({ time, root }) {
        const isServerClock = typeof time === 'number';
        const doc = isServerClock ? model_1.Model.withServerClock(time) : model_1.Model.withLogicalClock(this.decodeClock(time));
        this.decodeRoot(doc, root);
        return doc;
    }
    decodeClock(timestamps) {
        const [ts] = timestamps;
        const vectorClock = new clock_1.LogicalVectorClock(ts[0], ts[1]);
        const length = timestamps.length;
        for (let i = 0; i < length; i++) {
            const ts = timestamps[i];
            const [sessionId, time] = ts;
            vectorClock.observe(new clock_1.LogicalTimestamp(sessionId, time), 1);
        }
        return vectorClock;
    }
    decodeTimestamp(ts) {
        const isServerClock = typeof ts === 'number';
        return isServerClock ? new clock_1.ServerTimestamp(ts) : new clock_1.LogicalTimestamp(ts[0], ts[1]);
    }
    decodeRoot(doc, { id, node }) {
        const ts = this.decodeTimestamp(id);
        const jsonNode = node ? this.decodeNode(doc, node) : null;
        const root = new DocRootType_1.DocRootType(doc, ts, jsonNode);
        doc.root = root;
    }
    decodeNode(doc, node) {
        switch (node.type) {
            case 'obj':
                return this.decodeObj(doc, node);
            case 'arr':
                return this.decodeArr(doc, node);
            case 'str':
                return this.decodeStr(doc, node);
            case 'val':
                return this.decodeVal(doc, node);
            case 'const':
                return this.decodeConst(doc, node);
            case 'bin':
                return this.decodeBin(doc, node);
        }
        throw new Error('UNKNOWN_NODE');
    }
    decodeObj(doc, node) {
        const id = this.decodeTimestamp(node.id);
        const obj = new ObjectType_1.ObjectType(doc, id);
        const keys = Object.keys(node.chunks);
        for (const key of keys) {
            const val = node.chunks[key];
            obj.putChunk(key, new ObjectChunk_1.ObjectChunk(this.decodeTimestamp(val.id), this.decodeNode(doc, val.node)));
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeArr(doc, node) {
        const obj = new ArrayType_1.ArrayType(doc, this.decodeTimestamp(node.id));
        for (const c of node.chunks)
            obj.append(this.decodeArrChunk(doc, c));
        doc.nodes.index(obj);
        return obj;
    }
    decodeArrChunk(doc, c) {
        const id = this.decodeTimestamp(c.id);
        if (typeof c.span === 'number') {
            const chunk = new ArrayChunk_1.ArrayChunk(id, undefined);
            chunk.deleted = c.span;
            return chunk;
        }
        else
            return new ArrayChunk_1.ArrayChunk(id, c.nodes.map((n) => this.decodeNode(doc, n)));
    }
    decodeStr(doc, node) {
        const obj = new StringType_1.StringType(doc, this.decodeTimestamp(node.id));
        for (const c of node.chunks)
            obj.append(this.decodeStrChunk(doc, c));
        doc.nodes.index(obj);
        return obj;
    }
    decodeStrChunk(doc, c) {
        const id = this.decodeTimestamp(c.id);
        if (typeof c.span === 'number') {
            const chunk = new StringChunk_1.StringChunk(id, undefined);
            chunk.deleted = c.span;
            return chunk;
        }
        else
            return new StringChunk_1.StringChunk(id, c.value);
    }
    decodeBin(doc, node) {
        const obj = new BinaryType_1.BinaryType(doc, this.decodeTimestamp(node.id));
        for (const c of node.chunks)
            obj.append(this.decodeBinChunk(doc, c));
        doc.nodes.index(obj);
        return obj;
    }
    decodeBinChunk(doc, c) {
        const id = this.decodeTimestamp(c.id);
        if (typeof c.span === 'number') {
            const chunk = new BinaryChunk_1.BinaryChunk(id, undefined);
            chunk.deleted = c.span;
            return chunk;
        }
        else {
            const data = (0, decode_1.fromBase64)(c.value);
            return new BinaryChunk_1.BinaryChunk(id, data);
        }
    }
    decodeVal(doc, node) {
        const obj = new ValueType_1.ValueType(this.decodeTimestamp(node.id), this.decodeTimestamp(node.writeId), node.value);
        doc.nodes.index(obj);
        return obj;
    }
    decodeConst(doc, node) {
        switch (node.value) {
            case null:
                return constants_1.NULL;
            case true:
                return constants_1.TRUE;
            case false:
                return constants_1.FALSE;
            case undefined:
                return constants_1.UNDEFINED;
        }
        return new ConstantType_1.ConstantType(constants_2.ORIGIN, node.value);
    }
}
exports.Decoder = Decoder;
