"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcClient = void 0;
const tslib_1 = require("tslib");
const rxjs_1 = require("rxjs");
const nominal_1 = require("../messages/nominal");
const subscribeCompleteObserver_1 = require("../util/subscribeCompleteObserver");
const TimedQueue_1 = require("../util/TimedQueue");
class RpcClient {
    constructor({ send, bufferSize = 100, bufferTime = 10 }) {
        this.id = 1;
        this.calls = new Map();
        this.buffer = new TimedQueue_1.TimedQueue();
        this.buffer.itemLimit = bufferSize;
        this.buffer.timeLimit = bufferTime;
        this.buffer.onFlush = send;
    }
    getInflightCallCount() {
        return this.calls.size;
    }
    onMessages(messages) {
        const length = messages.length;
        for (let i = 0; i < length; i++)
            this.onMessage(messages[i]);
    }
    onMessage(message) {
        if (message instanceof nominal_1.ResponseCompleteMessage)
            return this.onResponseComplete(message);
        else if (message instanceof nominal_1.ResponseDataMessage)
            return this.onResponseData(message);
        else if (message instanceof nominal_1.ResponseErrorMessage)
            return this.onResponseError(message);
        return this.onRequestUnsubscribe(message);
    }
    onResponseComplete(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.resFinalized = true;
        if (data !== undefined)
            call.res$.next(data);
        call.res$.complete();
    }
    onResponseData(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.res$.next(data);
    }
    onResponseError(message) {
        const { id, data } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.resFinalized = true;
        call.res$.error(data);
    }
    onRequestUnsubscribe(message) {
        const { id } = message;
        const call = this.calls.get(id);
        if (!call)
            return;
        call.req$.complete();
    }
    call$(method, data) {
        const id = this.id++;
        if (this.id >= 0xffff)
            this.id = 1;
        if (this.calls.has(id))
            return this.call$(method, data);
        const req$ = new rxjs_1.Subject();
        const res$ = new rxjs_1.Subject();
        let finalizedStreams = 0;
        const cleanup = () => {
            finalizedStreams++;
            if (finalizedStreams === 2)
                this.calls.delete(id);
        };
        res$.subscribe({ error: cleanup, complete: cleanup });
        const entry = { req$, res$ };
        this.calls.set(id, entry);
        if ((0, rxjs_1.isObservable)(data)) {
            let firstMessageSent = false;
            (0, subscribeCompleteObserver_1.subscribeCompleteObserver)(req$, {
                next: (value) => {
                    const messageMethod = firstMessageSent ? '' : method;
                    firstMessageSent = true;
                    const message = new nominal_1.RequestDataMessage(id, messageMethod, value);
                    this.buffer.push(message);
                },
                error: (error) => {
                    cleanup();
                    const messageMethod = firstMessageSent ? '' : method;
                    const message = new nominal_1.RequestErrorMessage(id, messageMethod, error);
                    this.buffer.push(message);
                },
                complete: (value) => {
                    cleanup();
                    const messageMethod = firstMessageSent ? '' : method;
                    const message = new nominal_1.RequestCompleteMessage(id, messageMethod, value);
                    this.buffer.push(message);
                },
            });
            data.subscribe(req$);
        }
        else {
            this.buffer.push(new nominal_1.RequestCompleteMessage(id, method, data));
            req$.complete();
            cleanup();
        }
        return new rxjs_1.Observable((observer) => {
            res$.subscribe(observer);
            return () => {
                if (!entry.resFinalized)
                    this.buffer.push(new nominal_1.ResponseUnsubscribeMessage(id));
                res$.complete();
            };
        });
    }
    call(method, request) {
        return (0, tslib_1.__awaiter)(this, void 0, void 0, function* () {
            return yield (0, rxjs_1.firstValueFrom)(this.call$(method, request));
        });
    }
    notify(method, data) {
        this.buffer.push(new nominal_1.NotificationMessage(method, data));
    }
    stop(reason = 'STOP') {
        this.buffer.onFlush = (message) => { };
        for (const call of this.calls.values()) {
            call.req$.error(new Error(reason));
            call.req$.error(new Error(reason));
        }
        this.calls.clear();
    }
    disconnect() {
        this.stop('DISCONNECT');
    }
}
exports.RpcClient = RpcClient;
