import { ArrayChunk } from '../../types/rga-array/ArrayChunk';
import { ArrayType } from '../../types/rga-array/ArrayType';
import { BinaryChunk } from '../../types/rga-binary/BinaryChunk';
import { BinaryType } from '../../types/rga-binary/BinaryType';
import { ClockDecoder } from '../../../json-crdt-patch/codec/clock/ClockDecoder';
import { ConstantType } from '../../types/const/ConstantType';
import { DocRootType } from '../../types/lww-doc-root/DocRootType';
import { FALSE, NULL, TRUE, UNDEFINED } from '../../constants';
import { fromBase64 } from '../../../util/base64/decode';
import { LogicalTimestamp, ServerTimestamp } from '../../../json-crdt-patch/clock';
import { Model } from '../../model';
import { ObjectChunk } from '../../types/lww-object/ObjectChunk';
import { ObjectType } from '../../types/lww-object/ObjectType';
import { FALSE_ID, NULL_ID, ORIGIN, TRUE_ID, UNDEFINED_ID } from '../../../json-crdt-patch/constants';
import { StringChunk } from '../../types/rga-string/StringChunk';
import { StringType } from '../../types/rga-string/StringType';
import { ValueType } from '../../types/lww-value/ValueType';
export class Decoder {
    time;
    clockDecoder;
    decode(data) {
        const x = data[0];
        const isServerTime = typeof x === 'number';
        if (isServerTime) {
            this.time = x;
        }
        else {
            this.clockDecoder = ClockDecoder.fromArr(x);
        }
        const doc = isServerTime ? Model.withServerClock(x) : Model.withLogicalClock(this.clockDecoder.clock);
        const [rootId, index] = this.ts(data, 1);
        const rootNode = data[index] ? this.decodeNode(doc, data[index]) : null;
        doc.root = new DocRootType(doc, rootId, rootNode);
        return doc;
    }
    ts(arr, index) {
        const x = arr[index];
        if (typeof x === 'number') {
            if (x < 0) {
                const sessionIndex = -x;
                const timeDiff = arr[index + 1];
                return [this.clockDecoder.decodeId(sessionIndex, timeDiff), index + 2];
            }
            else {
                return [new ServerTimestamp(this.time - x), index + 1];
            }
        }
        else {
            const time = x[0];
            switch (time) {
                case ORIGIN.time:
                    return [ORIGIN, index + 1];
                case NULL_ID.time:
                    return [NULL_ID, index + 1];
                case TRUE_ID.time:
                    return [TRUE_ID, index + 1];
                case FALSE_ID.time:
                    return [FALSE_ID, index + 1];
                case UNDEFINED_ID.time:
                    return [UNDEFINED_ID, index + 1];
                default:
                    return [new LogicalTimestamp(0, time), index + 1];
            }
        }
    }
    decodeNode(doc, data) {
        switch (data) {
            case 1:
                return NULL;
            case 2:
                return FALSE;
            case 3:
                return TRUE;
            case 4:
                return UNDEFINED;
        }
        if (data instanceof Array) {
            switch (data[0]) {
                case 0:
                    return this.decodeObj(doc, data);
                case 1:
                    return this.decodeArr(doc, data);
                case 2:
                    return this.decodeStr(doc, data);
                case 3:
                    return this.decodeVal(doc, data);
                case 4:
                    return this.decodeConst(doc, data);
                case 5:
                    return this.decodeBin(doc, data);
            }
        }
        throw new Error('UNKNOWN_NODE');
    }
    decodeObj(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new ObjectType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const key = data[i];
            const [keyId, idx] = this.ts(data, i + 1);
            const node = this.decodeNode(doc, data[idx]);
            const chunk = new ObjectChunk(keyId, node);
            obj.putChunk(key, chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeArr(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new ArrayType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new ArrayChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                const nodes = content.map((c) => this.decodeNode(doc, c));
                chunk = new ArrayChunk(chunkId, nodes);
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeStr(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new StringType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new StringChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                chunk = new StringChunk(chunkId, content);
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeBin(doc, data) {
        const [id, index] = this.ts(data, 1);
        const obj = new BinaryType(doc, id);
        const length = data.length;
        for (let i = index; i < length;) {
            const [chunkId, idx] = this.ts(data, i);
            const content = data[idx];
            let chunk;
            if (typeof content === 'number') {
                chunk = new BinaryChunk(chunkId, undefined);
                chunk.deleted = content;
            }
            else {
                chunk = new BinaryChunk(chunkId, fromBase64(content));
            }
            obj.append(chunk);
            i = idx + 1;
        }
        doc.nodes.index(obj);
        return obj;
    }
    decodeVal(doc, data) {
        const [id, index1] = this.ts(data, 1);
        const [writeId, index2] = this.ts(data, index1);
        const value = data[index2];
        const obj = new ValueType(id, writeId, value);
        doc.nodes.index(obj);
        return obj;
    }
    decodeConst(doc, data) {
        const value = data[1];
        const obj = new ConstantType(ORIGIN, value);
        return obj;
    }
}
