import { ArrayType } from '../types/rga-array/ArrayType';
import { BinaryType } from '../types/rga-binary/BinaryType';
import { DeleteOperation } from '../../json-crdt-patch/operations/DeleteOperation';
import { DocRootType } from '../types/lww-doc-root/DocRootType';
import { FALSE, NULL, TRUE, UNDEFINED } from '../constants';
import { InsertArrayElementsOperation } from '../../json-crdt-patch/operations/InsertArrayElementsOperation';
import { InsertBinaryDataOperation } from '../../json-crdt-patch/operations/InsertBinaryDataOperation';
import { InsertStringSubstringOperation } from '../../json-crdt-patch/operations/InsertStringSubstringOperation';
import { LogicalVectorClock, ServerVectorClock } from '../../json-crdt-patch/clock';
import { LogicalNodeIndex, ServerNodeIndex } from './nodes';
import { MakeArrayOperation } from '../../json-crdt-patch/operations/MakeArrayOperation';
import { MakeBinaryOperation } from '../../json-crdt-patch/operations/MakeBinaryOperation';
import { MakeObjectOperation } from '../../json-crdt-patch/operations/MakeObjectOperation';
import { MakeStringOperation } from '../../json-crdt-patch/operations/MakeStringOperation';
import { MakeValueOperation } from '../../json-crdt-patch/operations/MakeValueOperation';
import { ModelApi } from './api/ModelApi';
import { ObjectType } from '../types/lww-object/ObjectType';
import { ORIGIN } from '../../json-crdt-patch/constants';
import { randomSessionId } from './util';
import { SetObjectKeysOperation } from '../../json-crdt-patch/operations/SetObjectKeysOperation';
import { SetRootOperation } from '../../json-crdt-patch/operations/SetRootOperation';
import { SetValueOperation } from '../../json-crdt-patch/operations/SetValueOperation';
import { StringType } from '../types/rga-string/StringType';
import { ValueType } from '../types/lww-value/ValueType';
export class Model {
    static withLogicalClock(clock) {
        clock = clock || new LogicalVectorClock(randomSessionId(), 0);
        const nodes = new LogicalNodeIndex();
        return new Model(clock, nodes);
    }
    static withServerClock(time) {
        const clock = new ServerVectorClock(time || 0);
        const nodes = new ServerNodeIndex();
        return new Model(clock, nodes);
    }
    root = new DocRootType(this, ORIGIN, null);
    clock;
    nodes;
    api;
    constructor(clock, nodes) {
        this.clock = clock;
        this.nodes = nodes;
        this.api = new ModelApi(this);
    }
    node(id) {
        if (id.getSessionId() === 0) {
            switch (id.time) {
                case 1:
                    return NULL;
                case 2:
                    return TRUE;
                case 3:
                    return FALSE;
                case 4:
                    return UNDEFINED;
            }
        }
        return this.nodes.get(id);
    }
    applyBatch(batch) {
        const patches = batch.patches;
        const { length } = patches;
        for (let i = 0; i < length; i++)
            this.applyPatch(patches[i]);
    }
    applyPatch(patch) {
        const ops = patch.ops;
        const { length } = ops;
        for (let i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    }
    applyOperation(op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation) {
            const oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation) {
            const obj = this.nodes.get(op.object);
            if (obj instanceof ObjectType)
                obj.insert(op);
        }
        else if (op instanceof SetValueOperation) {
            const obj = this.nodes.get(op.obj);
            if (obj instanceof ValueType)
                obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation) {
            const arr = this.nodes.get(op.arr);
            if (arr instanceof ArrayType)
                arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation) {
            const arr = this.nodes.get(op.obj);
            if (arr instanceof StringType)
                arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation) {
            const node = this.nodes.get(op.obj);
            if (node instanceof ArrayType)
                node.delete(op);
            else if (node instanceof StringType)
                node.onDelete(op);
            else if (node instanceof BinaryType)
                node.onDelete(op);
        }
        else if (op instanceof MakeBinaryOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new BinaryType(this, op.id));
        }
        else if (op instanceof InsertBinaryDataOperation) {
            const arr = this.nodes.get(op.obj);
            if (arr instanceof BinaryType)
                arr.onInsert(op);
        }
    }
    deleteNodeTree(value) {
        const isSystemNode = value.getSessionId() < 1;
        if (isSystemNode)
            return;
        const node = this.nodes.get(value);
        if (!node)
            return;
        for (const child of node.children())
            this.deleteNodeTree(child);
        this.nodes.delete(value);
    }
    advanceClocks() {
        for (const node of this.nodes.iterate()) {
            this.clock.observe(node.id, 1);
            if (node instanceof ObjectType) {
                for (const chunk of node.latest.values())
                    this.clock.observe(chunk.id, 1);
            }
            else if (node instanceof ArrayType) {
                for (const chunk of node.chunks())
                    this.clock.observe(chunk.id, chunk.span());
            }
            else if (node instanceof StringType) {
                for (const chunk of node.chunks())
                    this.clock.observe(chunk.id, chunk.span());
            }
            else if (node instanceof BinaryType) {
                for (const chunk of node.chunks())
                    this.clock.observe(chunk.id, chunk.span());
            }
            else if (node instanceof ValueType) {
                this.clock.observe(node.id, 1);
            }
        }
    }
    fork(sessionId = randomSessionId()) {
        const model = this.clock instanceof LogicalVectorClock
            ? Model.withLogicalClock(this.clock.fork(sessionId))
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        model.api.batch = this.api.batch.clone();
        return model;
    }
    clone() {
        return this.fork(this.clock.getSessionId());
    }
    toView() {
        return this.root.toJson();
    }
    toString(tab) {
        return this.root.toString(tab);
    }
}
