import { encoder } from '../json-pack/util';
import { JsExpression } from '../util/codegen/util/JsExpression';
import { normalizeAccessor } from '../util/codegen/util/normalizeAccessor';
import { Codegen, CodegenStepExecJs } from '../util/codegen';
const UINTS = ['u', 'u8', 'u16', 'u32', 'u64'];
const INTS = ['i', 'i8', 'i16', 'i32', 'i64', ...UINTS];
const join = (a, b) => {
    const res = new Uint8Array(a.length + b.length);
    res.set(a);
    res.set(b, a.length);
    return res;
};
class WriteBlobStep {
    arr;
    constructor(arr) {
        this.arr = arr;
    }
}
export class MsgPackSerializerCodegen {
    codegen;
    options;
    constructor(opts) {
        this.options = {
            ref: (id) => undefined,
            ...opts,
        };
        this.codegen = new Codegen({
            name: 'toMsgPack',
            args: 'r0',
            prologue: 'e.reset();',
            epilogue: 'return e.flush();',
            processSteps: (steps) => {
                const stepsJoined = [];
                for (let i = 0; i < steps.length; i++) {
                    const step = steps[i];
                    if (step instanceof CodegenStepExecJs)
                        stepsJoined.push(step);
                    else if (step instanceof WriteBlobStep) {
                        const last = stepsJoined[stepsJoined.length - 1];
                        if (last instanceof WriteBlobStep)
                            last.arr = join(last.arr, step.arr);
                        else
                            stepsJoined.push(step);
                    }
                }
                const execSteps = [];
                for (const step of stepsJoined) {
                    if (step instanceof CodegenStepExecJs) {
                        execSteps.push(step);
                    }
                    else if (step instanceof WriteBlobStep) {
                        execSteps.push(this.codegenBlob(step));
                    }
                }
                return execSteps;
            },
        });
        this.codegen.linkDependency(this.options.encoder, 'e');
    }
    execJs(js) {
        this.codegen.js(js);
    }
    writeBlob(arr) {
        this.codegen.step(new WriteBlobStep(arr));
    }
    getBlob(callback) {
        encoder.reset();
        callback(encoder);
        return encoder.flush();
    }
    registerCounter = 0;
    getRegister() {
        return `r${this.registerCounter++}`;
    }
    genAndWriteBlob(callback) {
        this.writeBlob(this.getBlob(callback));
    }
    onString(str, value) {
        if (str.const !== undefined) {
            this.writeBlob(encoder.encode(str.const));
            return;
        }
        if (str.format === 'ascii') {
            this.execJs(`e.encodeAsciiString(${value.use()});`);
            return;
        }
        this.execJs(`e.encodeString(${value.use()});`);
    }
    onNumber(num, value) {
        if (num.const !== undefined) {
            this.writeBlob(encoder.encode(num.const));
            return;
        }
        const isInteger = INTS.indexOf(num.format) > -1;
        if (isInteger) {
            const isUnsignedInteger = UINTS.indexOf(num.format) > -1;
            if (isUnsignedInteger) {
                this.execJs(`e.encodeUnsignedInteger(${value.use()});`);
            }
            else {
                this.execJs(`e.encodeInteger(${value.use()});`);
            }
            return;
        }
        this.execJs(`e.encodeNumber(${value.use()});`);
    }
    onBoolean(bool, value) {
        if (bool.const !== undefined) {
            this.writeBlob(encoder.encode(bool.const));
            return;
        }
        this.execJs(`e.encodeBoolean(${value.use()});`);
    }
    onNull() {
        this.execJs(`e.encodeNull();`);
    }
    onArray(arr, value) {
        if (arr.const !== undefined) {
            this.genAndWriteBlob((encoder) => encoder.encodeArray(arr.const));
            return;
        }
        const r = this.getRegister();
        const rl = this.getRegister();
        const ri = this.getRegister();
        const rItem = this.getRegister();
        this.execJs(`var ${r} = ${value.use()}, ${rl} = ${r}.length, ${ri} = 0, ${rItem};`);
        this.execJs(`e.encodeArrayHeader(${rl});`);
        this.execJs(`for(; ${ri} < ${rl}; ${ri}++) ` + '{');
        this.execJs(`${rItem} = ${r}[${ri}];`);
        this.onType(arr.type, new JsExpression(() => `${rItem}`));
        this.execJs(`}`);
    }
    onObject(obj, value) {
        if (obj.unknownFields) {
            this.execJs(`e.encodeObject(${value.use()});`);
            return;
        }
        let hasOptionalFields = false;
        for (let i = 0; i < obj.fields.length; i++) {
            if (obj.fields[i].isOptional) {
                hasOptionalFields = true;
                break;
            }
        }
        if (hasOptionalFields)
            this.onObjectVariableLength(obj, value);
        else
            this.onObjectFixedLength(obj, value);
    }
    onObjectVariableLength(obj, value) {
        const requiredFields = [];
        const optionalFields = [];
        for (let i = 0; i < obj.fields.length; i++) {
            const field = obj.fields[i];
            if (field.isOptional)
                optionalFields.push(field);
            else
                requiredFields.push(field);
        }
        const r = this.getRegister();
        const rk = this.getRegister();
        const rv = this.getRegister();
        this.execJs(`var ${r} = ${value.use()};`);
        this.execJs(`var ${rk} = Object.keys(${r});`);
        this.execJs(`e.encodeObjectHeader(${rk}.length);`);
        this.onRequiredFields(requiredFields, value, r);
        for (let i = 0; i < optionalFields.length; i++) {
            const field = optionalFields[i];
            const accessor = normalizeAccessor(field.key);
            this.execJs(`var ${rv} = ${r}${accessor};`);
            this.execJs(`if (${rv} !== undefined) {`);
            this.genAndWriteBlob((encoder) => encoder.encodeString(field.key));
            this.onType(field.type, new JsExpression(() => rv));
            this.execJs('}');
        }
    }
    onObjectFixedLength(obj, value) {
        const length = obj.fields.length;
        this.genAndWriteBlob((encoder) => encoder.encodeObjectHeader(length));
        const r = this.getRegister();
        value.addListener((expr) => {
            this.execJs(`var ${r} = ${expr};`);
        });
        this.onRequiredFields(obj.fields, value, r);
    }
    onRequiredFields(requiredFields, value, r) {
        const length = requiredFields.length;
        for (let i = 0; i < length; i++) {
            const field = requiredFields[i];
            this.genAndWriteBlob((encoder) => encoder.encodeString(field.key));
            const accessor = normalizeAccessor(field.key);
            this.onType(field.type, value.chain(() => `${r}${accessor}`));
        }
    }
    onBinary(value) {
        this.execJs(`e.encodeBinary(${value.use()});`);
    }
    onRef(ref, value) {
        const type = this.options.ref(ref.ref);
        switch (typeof type) {
            case 'function': {
                const d = this.codegen.linkDependency(type);
                this.execJs(`${d}(${value.use()}, e);`);
                break;
            }
            case 'object': {
                this.onType(type, value);
                break;
            }
            default:
                throw new Error(`Unknown [ref = ${ref.ref}].`);
        }
    }
    onAny(value) {
        this.execJs(`e.encodeAny(${value.use()});`);
    }
    onType(type, value) {
        switch (type.__t) {
            case 'str': {
                this.onString(type, value);
                break;
            }
            case 'num': {
                this.onNumber(type, value);
                break;
            }
            case 'bool': {
                this.onBoolean(type, value);
                break;
            }
            case 'nil': {
                this.onNull();
                break;
            }
            case 'arr': {
                this.onArray(type, value);
                break;
            }
            case 'obj': {
                this.onObject(type, value);
                break;
            }
            case 'bin': {
                this.onBinary(value);
                break;
            }
            case 'ref': {
                this.onRef(type, value);
                break;
            }
            default: {
                this.onAny(value);
                break;
            }
        }
    }
    codegenBlob(step) {
        const lines = [];
        const ro = this.getRegister();
        const ru = this.getRegister();
        lines.push(`var ${ro} = e.offset, ${ru} = e.uint8;`);
        lines.push(`e.ensureCapacity(${step.arr.length});`);
        lines.push(`e.offset += ${step.arr.length};`);
        for (let i = 0; i < step.arr.length; i++)
            lines.push(`${ru}[${ro} + ${i}] = ${step.arr[i]};`);
        const js = lines.join('\n');
        return new CodegenStepExecJs(js);
    }
    run() {
        const r = this.getRegister();
        const value = new JsExpression(() => r);
        this.onType(this.options.type, value);
        return this;
    }
    generate() {
        return this.codegen.generate();
    }
    compile() {
        return this.codegen.compile();
    }
    compilePartial() {
        return this.codegen.compile({
            name: 'partial',
            args: 'r0, e',
            epilogue: '',
            prologue: '',
        });
    }
}
