import { createBoolValidator, createObjValidator, } from '../json-type-validator';
import { dynamicFunction } from '../util/codegen/dynamicFunction';
import { JsonSerializerCodegen, MsgPackSerializerCodegen, } from '../json-type-serializer';
import { encoderFull } from '../json-pack/util';
import { toJsonSchema } from '../json-type-schema';
import { toTypeScriptText } from '../json-type-typescript/toTypeScriptText';
import { exportDeclaration } from '../json-type-typescript/export';
const literalTypes = ['bin', 'bool', 'nil', 'num', 'str'];
export class JsonTypeSystem {
    options;
    constructor(options) {
        this.options = options;
    }
    ref = (ref) => {
        const type = this.options.types[ref];
        if (!type)
            throw new Error(`Type [ref = ${ref}] not found.`);
        return type;
    };
    hasType(ref) {
        return Object.prototype.hasOwnProperty.call(this.options.types, ref);
    }
    fastValidatorCache = {};
    fastValidatorUsage = {};
    getFastValidator(ref) {
        const type = this.ref(ref);
        const cachedFastValidator = this.fastValidatorCache[ref];
        if (cachedFastValidator) {
            this.fastValidatorUsage[ref] = (this.fastValidatorUsage[ref] || 0) + 1;
            return cachedFastValidator;
        }
        const [validator, setValidator] = dynamicFunction(() => {
            throw new Error(`Type [ref = ${ref}] not implemented.`);
        });
        this.fastValidatorCache[ref] = validator;
        const realValidator = createBoolValidator(type, {
            customValidators: this.options.customValidators,
            ref: (id) => this.getFastValidator(id),
            skipObjectExtraFieldsCheck: true,
            unsafeMode: true,
        });
        if (this.fastValidatorUsage[ref])
            setValidator(realValidator);
        this.fastValidatorCache[ref] = realValidator;
        return realValidator;
    }
    fullValidatorCache = {};
    getFullValidator(ref) {
        const type = this.ref(ref);
        const cachedFullValidator = this.fullValidatorCache[ref];
        if (cachedFullValidator)
            return cachedFullValidator;
        const [validator, setValidator] = dynamicFunction(() => {
            throw new Error(`Type [ref = ${ref}] not implemented.`);
        });
        this.fullValidatorCache[ref] = validator;
        setValidator(createObjValidator(type, {
            customValidators: this.options.customValidators,
            ref: (id) => this.getFullValidator(id),
        }));
        return validator;
    }
    jsonSerializerCache = {};
    jsonSerializerUsage = {};
    getJsonSerializer(ref) {
        const type = this.ref(ref);
        const jsonSerializerCached = this.jsonSerializerCache[ref];
        if (jsonSerializerCached) {
            this.jsonSerializerUsage[ref] = (this.jsonSerializerUsage[ref] || 0) + 1;
            return jsonSerializerCached;
        }
        const codegen = new JsonSerializerCodegen({
            type,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getJsonSerializer(id);
            },
        });
        const [serializer, setSerializer] = dynamicFunction(() => {
            throw new Error(`Type [ref = ${ref}] JSON serializer not implemented.`);
        });
        this.jsonSerializerCache[ref] = serializer;
        const realSerializer = codegen.run().compile();
        if (this.jsonSerializerUsage[ref])
            setSerializer(realSerializer);
        this.jsonSerializerCache[ref] = realSerializer;
        return realSerializer;
    }
    msgPackPartialEncoderCache = {};
    msgPackPartialEncoderUsage = {};
    getMsgPackPartialEncoder(ref) {
        const type = this.ref(ref);
        const msgPackPartialEncoderCached = this.msgPackPartialEncoderCache[ref];
        if (msgPackPartialEncoderCached) {
            this.msgPackPartialEncoderUsage[ref] = (this.msgPackPartialEncoderUsage[ref] || 0) + 1;
            return msgPackPartialEncoderCached;
        }
        const codegen = new MsgPackSerializerCodegen({
            type,
            encoder: encoderFull,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getMsgPackPartialEncoder(id);
            },
        });
        const [partialEncoder, setPartialEncoder] = dynamicFunction(() => {
            throw new Error(`Type [ref = ${ref}] MessagePack partial encoder not implemented.`);
        });
        this.msgPackPartialEncoderCache[ref] = partialEncoder;
        const realPartialEncoder = codegen.run().compilePartial();
        if (this.msgPackPartialEncoderUsage[ref])
            setPartialEncoder(realPartialEncoder);
        this.msgPackPartialEncoderCache[ref] = realPartialEncoder;
        return realPartialEncoder;
    }
    msgPackEncoderCache = {};
    getMsgPackEncoder(ref) {
        const type = this.ref(ref);
        const msgPackSerializerCached = this.msgPackEncoderCache[ref];
        if (msgPackSerializerCached)
            return msgPackSerializerCached;
        const codegen = new MsgPackSerializerCodegen({
            type,
            encoder: encoderFull,
            ref: (id) => {
                const type = this.ref(id);
                const isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : this.getMsgPackPartialEncoder(id);
            },
        });
        return (this.msgPackEncoderCache[ref] = codegen.run().compile());
    }
    toJsonSchema(ref, withDefs) {
        const type = this.ref(ref);
        const defs = withDefs ? {} : undefined;
        let schema = toJsonSchema(type, {
            ref: this.ref,
            defs,
        });
        if (defs) {
            schema = {
                $id: ref,
                ...schema,
                $defs: defs,
            };
        }
        return schema;
    }
    toTypeScriptText(ref) {
        const type = this.ref(ref);
        return toTypeScriptText(type, '');
    }
    toTsAst(ref) {
        const statements = [];
        if (ref) {
            exportDeclaration(ref, {
                ref: this.ref,
                statements,
            });
        }
        else {
            const types = this.options.types;
            for (const ref of Object.keys(types)) {
                exportDeclaration(ref, {
                    ref: this.ref,
                    statements,
                });
            }
        }
        return statements;
    }
}
