const parseIdentifier = (id) => {
    const parts = id.split('.');
    return {
        name: parts.pop(),
        namespaces: parts,
    };
};
const augmentWithComment = (type, node) => {
    if (type.title || type.description) {
        let comment = '';
        if (type.title)
            comment += '# ' + type.title;
        if (type.title && type.description)
            comment += '\n\n';
        if (type.description)
            comment += type.description;
        node.comment = comment;
    }
};
export const exportDeclaration = (ref, ctx) => {
    const identifier = parseIdentifier(ref);
    let module = {
        node: 'ModuleDeclaration',
        name: '',
        statements: ctx.statements,
    };
    for (const namespace of identifier.namespaces) {
        let innerModule = module.statements.find((s) => s.node === 'ModuleDeclaration' && s.name === namespace);
        if (!innerModule) {
            innerModule = {
                node: 'ModuleDeclaration',
                name: namespace,
                statements: [],
            };
            module.statements.push(innerModule);
        }
        module = innerModule;
    }
    const exists = module.statements.find((s) => s.name === identifier.name);
    if (exists)
        return;
    const type = ctx.ref(ref);
    switch (type.__t) {
        case 'obj': {
            const node = {
                node: 'InterfaceDeclaration',
                name: identifier.name,
                members: [],
            };
            module.statements.push(node);
            for (const field of type.fields) {
                const member = {
                    node: 'PropertySignature',
                    name: field.key,
                    type: toTsType(field.type, ctx),
                };
                if (field.isOptional)
                    member.optional = true;
                augmentWithComment(field, member);
                node.members.push(member);
            }
            if (type.unknownFields)
                node.members.push({
                    node: 'IndexSignature',
                    type: { node: 'UnknownKeyword' },
                });
            augmentWithComment(type, node);
            return;
        }
        default: {
            const node = {
                node: 'TypeAliasDeclaration',
                name: identifier.name,
                type: toTsType(type, ctx),
            };
            augmentWithComment(type, node);
            module.statements.push(node);
            return;
        }
    }
};
const toTsType = (type, ctx) => {
    switch (type.__t) {
        case 'str': {
            const node = typeof type.const === 'string' ? { node: 'StringLiteral', text: type.const } : { node: 'StringKeyword' };
            return node;
        }
        case 'num': {
            const node = typeof type.const === 'number'
                ? { node: 'NumericLiteral', text: JSON.stringify(type.const) }
                : { node: 'NumberKeyword' };
            return node;
        }
        case 'bool': {
            const node = typeof type.const === 'boolean'
                ? type.const
                    ? { node: 'TrueKeyword' }
                    : { node: 'FalseKeyword' }
                : { node: 'BooleanKeyword' };
            return node;
        }
        case 'nil': {
            const node = { node: 'NullKeyword' };
            return node;
        }
        case 'arr': {
            const node = {
                node: 'ArrayType',
                elementType: toTsType(type.type, ctx),
            };
            return node;
        }
        case 'obj': {
            const node = {
                node: 'TypeLiteral',
                members: [],
            };
            for (const field of type.fields) {
                const member = {
                    node: 'PropertySignature',
                    name: field.key,
                    type: toTsType(field.type, ctx),
                };
                if (field.isOptional)
                    member.optional = true;
                augmentWithComment(field, member);
                node.members.push(member);
            }
            if (type.unknownFields)
                node.members.push({
                    node: 'IndexSignature',
                    type: { node: 'UnknownKeyword' },
                });
            augmentWithComment(type, node);
            return node;
        }
        case 'enum': {
            const node = {
                node: 'UnionType',
                types: type.values.map(toTsLiteral),
            };
            return node;
        }
        case 'or': {
            const node = {
                node: 'UnionType',
                types: type.types.map((t) => toTsType(t, ctx)),
            };
            return node;
        }
        case 'ref': {
            const node = {
                node: 'TypeReference',
                typeName: type.ref,
            };
            exportDeclaration(type.ref, ctx);
            return node;
        }
        default: {
            const node = { node: 'UnknownKeyword' };
            return node;
        }
    }
};
const toTsLiteral = (value) => {
    switch (typeof value) {
        case 'string': {
            const node = { node: 'StringLiteral', text: value };
            return node;
        }
        case 'number': {
            const node = { node: 'NumericLiteral', text: value.toString() };
            return node;
        }
        case 'boolean': {
            const node = { node: value ? 'TrueKeyword' : 'FalseKeyword' };
            return node;
        }
        default: {
            const node = { node: 'UnknownKeyword' };
            return node;
        }
    }
};
