import { NotificationMessage, RequestCompleteMessage, RequestDataMessage, RequestErrorMessage, RequestUnsubscribeMessage, ResponseCompleteMessage, ResponseDataMessage, ResponseErrorMessage, ResponseUnsubscribeMessage, } from '../messages/nominal';
import { subscribeCompleteObserver } from '../util/subscribeCompleteObserver';
import { TimedQueue } from '../util/TimedQueue';
import { RpcServerError } from './constants';
import { ErrorLikeErrorFormatter, RpcError } from './error';
export { RpcServerError };
export class RpcServer {
    caller;
    error;
    activeStreamCalls = new Map();
    send;
    onSend;
    onNotification;
    constructor({ caller, error, send, onNotification: notify, bufferSize = 10, bufferTime = 1 }) {
        this.caller = caller;
        this.error = error || new ErrorLikeErrorFormatter();
        this.onNotification = notify;
        this.onSend = send;
        if (bufferTime) {
            const buffer = new TimedQueue();
            buffer.itemLimit = bufferSize;
            buffer.timeLimit = bufferTime;
            buffer.onFlush = (messages) => this.onSend(messages);
            this.send = (message) => {
                buffer.push(message);
            };
        }
        else {
            this.send = (message) => {
                this.onSend([message]);
            };
        }
    }
    notifMessage(method, data) {
        return new NotificationMessage(method, data);
    }
    resCompleteMessage(id, data) {
        return new ResponseCompleteMessage(id, data);
    }
    resDataMessage(id, data) {
        return new ResponseDataMessage(id, data);
    }
    resErrorMessage(id, data) {
        return new ResponseErrorMessage(id, data);
    }
    reqUnsubscribeMessage(id) {
        return new RequestUnsubscribeMessage(id);
    }
    createStaticCall(name, request, ctx) {
        return this.caller.call(name, request, ctx);
    }
    execStaticCall(id, name, request, ctx) {
        this.createStaticCall(name, request, ctx)
            .then((response) => {
            this.send(this.resCompleteMessage(id, response));
        })
            .catch((error) => {
            this.send(this.resErrorMessage(id, error));
        });
    }
    createCall(name, ctx) {
        const streamCallback = (method, ctx, req$) => method.call$(ctx, req$);
        return this.caller.createCall(name, ctx, (name, req, ctx) => this.caller.call(name, req, ctx), streamCallback);
    }
    onStreamError = (id, error) => {
        this.send(this.resErrorMessage(id, error));
        this.activeStreamCalls.delete(id);
    };
    onMessage(message, ctx) {
        try {
            if (message instanceof RequestDataMessage)
                this.onRequestDataMessage(message, ctx);
            else if (message instanceof RequestCompleteMessage)
                this.onRequestCompleteMessage(message, ctx);
            else if (message instanceof RequestErrorMessage)
                this.onRequestErrorMessage(message, ctx);
            else if (message instanceof NotificationMessage)
                this.onNotificationMessage(message, ctx);
            else if (message instanceof ResponseUnsubscribeMessage)
                this.onUnsubscribeMessage(message);
        }
        catch (error) {
            const formattedError = this.error.format(error);
            const message = this.notifMessage('.err', formattedError);
            this.send(message);
        }
    }
    onMessages(messages, ctx) {
        const length = messages.length;
        for (let i = 0; i < length; i++)
            this.onMessage(messages[i], ctx);
    }
    stop(reason = RpcServerError.Stop) {
        this.send = (() => { });
        this.onNotification = (name, data, ctx) => { };
        for (const call of this.activeStreamCalls.values())
            call.req$.error(new RpcError(reason));
        this.activeStreamCalls.clear();
    }
    disconnect() {
        this.stop(RpcServerError.Disconnect);
    }
    sendError(id, code) {
        const data = this.error.formatCode(code);
        const message = this.resErrorMessage(id, data);
        this.send(message);
    }
    createStreamCall(id, name, ctx) {
        const call = this.createCall(name, ctx);
        this.activeStreamCalls.set(id, call);
        subscribeCompleteObserver(call.res$, {
            next: (value) => {
                this.send(this.resDataMessage(id, value));
            },
            error: (error) => {
                this.onStreamError(id, error);
            },
            complete: (value) => {
                this.send(this.resCompleteMessage(id, value));
                this.activeStreamCalls.delete(id);
            },
        });
        call.reqUnsubscribe$.subscribe(() => {
            if (this.activeStreamCalls.has(id))
                this.send(this.reqUnsubscribeMessage(id));
        });
        return call;
    }
    onRequestDataMessage(message, ctx) {
        const { id, method, data } = message;
        let call = this.activeStreamCalls.get(id);
        if (!call) {
            if (!method) {
                this.sendError(id, RpcServerError.NoMethodSpecified);
                return;
            }
            call = this.createStreamCall(id, method, ctx);
        }
        if (call) {
            if (data !== undefined) {
                call.req$.next(data);
            }
        }
    }
    onRequestCompleteMessage(message, ctx) {
        const { id, method, data } = message;
        const call = this.activeStreamCalls.get(id);
        if (call) {
            const { req$ } = call;
            if (data !== undefined)
                req$.next(data);
            req$.complete();
            return;
        }
        if (!method) {
            this.sendError(id, RpcServerError.NoMethodSpecified);
            return;
        }
        if (!this.caller.exists(method)) {
            this.sendError(id, RpcServerError.MethodNotFound);
            return;
        }
        const rpcMethod = this.caller.get(method);
        if (rpcMethod.isStreaming) {
            const newCall = this.createStreamCall(id, method, ctx);
            if (newCall) {
                if (data !== undefined) {
                    newCall.req$.next(data);
                    newCall.req$.complete();
                }
            }
        }
        else
            this.execStaticCall(id, method, data, ctx);
    }
    onRequestErrorMessage(message, ctx) {
        const { id, method, data } = message;
        const call = this.activeStreamCalls.get(id);
        if (call) {
            return call.req$.error(data);
        }
        if (!method)
            return this.sendError(id, RpcServerError.NoMethodSpecified);
        if (!this.caller.exists(method))
            return this.sendError(id, RpcServerError.MethodNotFound);
        const rpcMethod = this.caller.get(method);
        if (!rpcMethod.isStreaming)
            return this.sendError(id, RpcServerError.ErrorForStaticMethod);
        const streamCall = this.createStreamCall(id, method, ctx);
        if (!streamCall)
            return;
        streamCall.req$.error(data);
    }
    onUnsubscribeMessage(message) {
        const { id } = message;
        const call = this.activeStreamCalls.get(id);
        if (!call)
            return;
        this.activeStreamCalls.delete(id);
        call.req$.complete();
    }
    onNotificationMessage(message, ctx) {
        const { method, data } = message;
        if (!method || method.length > 128)
            throw new RpcError(RpcServerError.InvalidNotificationName);
        this.onNotification(method, data, ctx);
    }
}
