"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryRxServer = void 0;
var tslib_1 = require("tslib");
var messages_1 = require("./messages");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var TimedQueue_1 = require("../json-rx/TimedQueue");
var codec_1 = require("./codec");
var util_1 = require("../json-rx/util");
var isUint8Array_1 = require("../util/isUint8Array");
var ERR_UNKNOWN = new Uint8Array([0]);
var ERR_ID_TAKEN = new Uint8Array([1]);
var ERR_TOO_MANY_SUBSCRIPTIONS = new Uint8Array([2]);
var ERR_INVALID_DATA = new Uint8Array([3]);
var BinaryRxServer = (function () {
    function BinaryRxServer(_a) {
        var _this = this;
        var send = _a.send, call = _a.call, notify = _a.notify, _b = _a.maxActiveSubscriptions, maxActiveSubscriptions = _b === void 0 ? 30 : _b, _c = _a.bufferSize, bufferSize = _c === void 0 ? 10 : _c, _d = _a.bufferTime, bufferTime = _d === void 0 ? 1 : _d;
        this.active = new Map();
        this.encoder = new codec_1.Encoder();
        this.call = call;
        this.notify = notify;
        this.maxActiveSubscriptions = maxActiveSubscriptions;
        if (bufferTime) {
            var buffer_1 = new TimedQueue_1.TimedQueue();
            buffer_1.itemLimit = bufferSize;
            buffer_1.timeLimit = bufferTime;
            buffer_1.onFlush = function (messages) {
                send(_this.encoder.encode(messages));
            };
            this.send = function (message) { return buffer_1.push(message); };
        }
        else {
            this.send = function (message) {
                send(_this.encoder.encode([message]));
            };
        }
    }
    BinaryRxServer.prototype.sendError = function (id, data) {
        this.send(new messages_1.ErrorMessage(id, data));
    };
    BinaryRxServer.prototype.onSubscribe = function (message, ctx) {
        var _this = this;
        var id = message.id, method = message.method, data = message.data;
        (0, util_1.assertName)(method);
        try {
            var active = this.active.get(id);
            if (active) {
                this.active.delete(id);
                active.unsubscribe();
                this.sendError(id, ERR_ID_TAKEN);
                return;
            }
            if (this.active.size >= this.maxActiveSubscriptions)
                return this.sendError(id, ERR_TOO_MANY_SUBSCRIPTIONS);
            var callResult = this.call(method, data, ctx);
            var observable = ((0, rxjs_1.isObservable)(callResult)
                ? callResult
                : (0, rxjs_1.from)(callResult).pipe((0, operators_1.mergeMap)(function (value) { return ((0, rxjs_1.isObservable)(value) ? value : (0, rxjs_1.of)(value)); })));
            var ref_1 = { buffer: [] };
            var done_1 = false;
            var subscription = observable.subscribe(function (data) {
                if (!(0, isUint8Array_1.isUint8Array)(data))
                    return _this.sendError(id, ERR_INVALID_DATA);
                if (data === undefined)
                    return;
                ref_1.buffer.push(data);
                (0, util_1.microtask)(function () {
                    var e_1, _a;
                    if (!ref_1.buffer.length)
                        return;
                    try {
                        try {
                            for (var _b = (0, tslib_1.__values)(ref_1.buffer), _c = _b.next(); !_c.done; _c = _b.next()) {
                                var payload = _c.value;
                                _this.send(new messages_1.DataMessage(id, payload));
                            }
                        }
                        catch (e_1_1) { e_1 = { error: e_1_1 }; }
                        finally {
                            try {
                                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                            }
                            finally { if (e_1) throw e_1.error; }
                        }
                    }
                    finally {
                        ref_1.buffer = [];
                    }
                });
            }, function (error) {
                done_1 = true;
                _this.active.delete(id);
                _this.sendError(id, (0, isUint8Array_1.isUint8Array)(error) ? error : ERR_UNKNOWN);
            }, function () {
                done_1 = true;
                _this.active.delete(id);
                try {
                    if (!ref_1.buffer.length)
                        return _this.send(new messages_1.CompleteMessage(id, undefined));
                    var last = ref_1.buffer.length - 1;
                    for (var i = 0; i <= last; i++) {
                        var isLast = i === last;
                        var message_1 = isLast
                            ? new messages_1.CompleteMessage(id, ref_1.buffer[i])
                            : new messages_1.DataMessage(id, ref_1.buffer[i]);
                        _this.send(message_1);
                    }
                }
                finally {
                    ref_1.buffer = [];
                }
            });
            if (!done_1)
                this.active.set(id, subscription);
        }
        catch (error) {
            this.sendError(id, Buffer.isBuffer(error) ? error : ERR_UNKNOWN);
        }
    };
    BinaryRxServer.prototype.onUnsubscribe = function (message) {
        var id = message.id;
        var subscription = this.active.get(id);
        if (!subscription)
            return;
        this.active.delete(id);
        subscription.unsubscribe();
    };
    BinaryRxServer.prototype.onNotification = function (message, ctx) {
        var method = message.method, data = message.data;
        (0, util_1.assertName)(method);
        this.notify(method, data, ctx);
    };
    BinaryRxServer.prototype.onArray = function (arr, ctx) {
        var messages = (0, codec_1.decodeFullMessages)(arr, 0);
        this.onMessages(messages, ctx);
    };
    BinaryRxServer.prototype.onMessages = function (messages, ctx) {
        var length = messages.length;
        for (var i = 0; i < length; i++) {
            var message = messages[i];
            if (message instanceof messages_1.NotificationMessage)
                this.onNotification(message, ctx);
            else if (message instanceof messages_1.SubscribeMessage)
                this.onSubscribe(message, ctx);
            else if (message instanceof messages_1.UnsubscribeMessage)
                this.onUnsubscribe(message);
        }
    };
    BinaryRxServer.prototype.stop = function () {
        var e_2, _a;
        this.send = function (message) { };
        try {
            for (var _b = (0, tslib_1.__values)(this.active.values()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var sub = _c.value;
                sub.unsubscribe();
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_2) throw e_2.error; }
        }
    };
    return BinaryRxServer;
}());
exports.BinaryRxServer = BinaryRxServer;
