"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = void 0;
var tslib_1 = require("tslib");
var ArrayType_1 = require("../types/rga-array/ArrayType");
var BinaryType_1 = require("../types/rga-binary/BinaryType");
var DeleteOperation_1 = require("../../json-crdt-patch/operations/DeleteOperation");
var DocRootType_1 = require("../types/lww-doc-root/DocRootType");
var constants_1 = require("../constants");
var InsertArrayElementsOperation_1 = require("../../json-crdt-patch/operations/InsertArrayElementsOperation");
var InsertBinaryDataOperation_1 = require("../../json-crdt-patch/operations/InsertBinaryDataOperation");
var InsertStringSubstringOperation_1 = require("../../json-crdt-patch/operations/InsertStringSubstringOperation");
var clock_1 = require("../../json-crdt-patch/clock");
var nodes_1 = require("./nodes");
var MakeArrayOperation_1 = require("../../json-crdt-patch/operations/MakeArrayOperation");
var MakeBinaryOperation_1 = require("../../json-crdt-patch/operations/MakeBinaryOperation");
var MakeObjectOperation_1 = require("../../json-crdt-patch/operations/MakeObjectOperation");
var MakeStringOperation_1 = require("../../json-crdt-patch/operations/MakeStringOperation");
var MakeValueOperation_1 = require("../../json-crdt-patch/operations/MakeValueOperation");
var ModelApi_1 = require("./api/ModelApi");
var ObjectType_1 = require("../types/lww-object/ObjectType");
var constants_2 = require("../../json-crdt-patch/constants");
var util_1 = require("./util");
var SetObjectKeysOperation_1 = require("../../json-crdt-patch/operations/SetObjectKeysOperation");
var SetRootOperation_1 = require("../../json-crdt-patch/operations/SetRootOperation");
var SetValueOperation_1 = require("../../json-crdt-patch/operations/SetValueOperation");
var StringType_1 = require("../types/rga-string/StringType");
var ValueType_1 = require("../types/lww-value/ValueType");
var Model = (function () {
    function Model(clock, nodes) {
        this.root = new DocRootType_1.DocRootType(this, constants_2.ORIGIN, null);
        this.clock = clock;
        this.nodes = nodes;
        this.api = new ModelApi_1.ModelApi(this);
    }
    Model.withLogicalClock = function (clock) {
        clock = clock || new clock_1.LogicalVectorClock((0, util_1.randomSessionId)(), 0);
        var nodes = new nodes_1.LogicalNodeIndex();
        return new Model(clock, nodes);
    };
    Model.withServerClock = function (time) {
        var clock = new clock_1.ServerVectorClock(time || 0);
        var nodes = new nodes_1.ServerNodeIndex();
        return new Model(clock, nodes);
    };
    Model.prototype.node = function (id) {
        if (id.getSessionId() === 0) {
            switch (id.time) {
                case 1:
                    return constants_1.NULL;
                case 2:
                    return constants_1.TRUE;
                case 3:
                    return constants_1.FALSE;
                case 4:
                    return constants_1.UNDEFINED;
            }
        }
        return this.nodes.get(id);
    };
    Model.prototype.applyBatch = function (batch) {
        var patches = batch.patches;
        var length = patches.length;
        for (var i = 0; i < length; i++)
            this.applyPatch(patches[i]);
    };
    Model.prototype.applyPatch = function (patch) {
        var ops = patch.ops;
        var length = ops.length;
        for (var i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    };
    Model.prototype.applyOperation = function (op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType_1.ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType_1.ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType_1.StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation_1.MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType_1.ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation_1.SetRootOperation) {
            var oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            var obj = this.nodes.get(op.object);
            if (obj instanceof ObjectType_1.ObjectType)
                obj.insert(op);
        }
        else if (op instanceof SetValueOperation_1.SetValueOperation) {
            var obj = this.nodes.get(op.obj);
            if (obj instanceof ValueType_1.ValueType)
                obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            var arr = this.nodes.get(op.arr);
            if (arr instanceof ArrayType_1.ArrayType)
                arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            var arr = this.nodes.get(op.obj);
            if (arr instanceof StringType_1.StringType)
                arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation_1.DeleteOperation) {
            var node = this.nodes.get(op.obj);
            if (node instanceof ArrayType_1.ArrayType)
                node.delete(op);
            else if (node instanceof StringType_1.StringType)
                node.onDelete(op);
            else if (node instanceof BinaryType_1.BinaryType)
                node.onDelete(op);
        }
        else if (op instanceof MakeBinaryOperation_1.MakeBinaryOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new BinaryType_1.BinaryType(this, op.id));
        }
        else if (op instanceof InsertBinaryDataOperation_1.InsertBinaryDataOperation) {
            var arr = this.nodes.get(op.obj);
            if (arr instanceof BinaryType_1.BinaryType)
                arr.onInsert(op);
        }
    };
    Model.prototype.deleteNodeTree = function (value) {
        var e_1, _a;
        var isSystemNode = value.getSessionId() < 1;
        if (isSystemNode)
            return;
        var node = this.nodes.get(value);
        if (!node)
            return;
        try {
            for (var _b = (0, tslib_1.__values)(node.children()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var child = _c.value;
                this.deleteNodeTree(child);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.nodes.delete(value);
    };
    Model.prototype.advanceClocks = function () {
        var e_2, _a, e_3, _b, e_4, _c, e_5, _d, e_6, _e;
        try {
            for (var _f = (0, tslib_1.__values)(this.nodes.iterate()), _g = _f.next(); !_g.done; _g = _f.next()) {
                var node = _g.value;
                this.clock.observe(node.id, 1);
                if (node instanceof ObjectType_1.ObjectType) {
                    try {
                        for (var _h = (e_3 = void 0, (0, tslib_1.__values)(node.latest.values())), _j = _h.next(); !_j.done; _j = _h.next()) {
                            var chunk = _j.value;
                            this.clock.observe(chunk.id, 1);
                        }
                    }
                    catch (e_3_1) { e_3 = { error: e_3_1 }; }
                    finally {
                        try {
                            if (_j && !_j.done && (_b = _h.return)) _b.call(_h);
                        }
                        finally { if (e_3) throw e_3.error; }
                    }
                }
                else if (node instanceof ArrayType_1.ArrayType) {
                    try {
                        for (var _k = (e_4 = void 0, (0, tslib_1.__values)(node.chunks())), _l = _k.next(); !_l.done; _l = _k.next()) {
                            var chunk = _l.value;
                            this.clock.observe(chunk.id, chunk.span());
                        }
                    }
                    catch (e_4_1) { e_4 = { error: e_4_1 }; }
                    finally {
                        try {
                            if (_l && !_l.done && (_c = _k.return)) _c.call(_k);
                        }
                        finally { if (e_4) throw e_4.error; }
                    }
                }
                else if (node instanceof StringType_1.StringType) {
                    try {
                        for (var _m = (e_5 = void 0, (0, tslib_1.__values)(node.chunks())), _o = _m.next(); !_o.done; _o = _m.next()) {
                            var chunk = _o.value;
                            this.clock.observe(chunk.id, chunk.span());
                        }
                    }
                    catch (e_5_1) { e_5 = { error: e_5_1 }; }
                    finally {
                        try {
                            if (_o && !_o.done && (_d = _m.return)) _d.call(_m);
                        }
                        finally { if (e_5) throw e_5.error; }
                    }
                }
                else if (node instanceof BinaryType_1.BinaryType) {
                    try {
                        for (var _p = (e_6 = void 0, (0, tslib_1.__values)(node.chunks())), _q = _p.next(); !_q.done; _q = _p.next()) {
                            var chunk = _q.value;
                            this.clock.observe(chunk.id, chunk.span());
                        }
                    }
                    catch (e_6_1) { e_6 = { error: e_6_1 }; }
                    finally {
                        try {
                            if (_q && !_q.done && (_e = _p.return)) _e.call(_p);
                        }
                        finally { if (e_6) throw e_6.error; }
                    }
                }
                else if (node instanceof ValueType_1.ValueType) {
                    this.clock.observe(node.id, 1);
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_g && !_g.done && (_a = _f.return)) _a.call(_f);
            }
            finally { if (e_2) throw e_2.error; }
        }
    };
    Model.prototype.fork = function (sessionId) {
        if (sessionId === void 0) { sessionId = (0, util_1.randomSessionId)(); }
        var model = this.clock instanceof clock_1.LogicalVectorClock
            ? Model.withLogicalClock(this.clock.fork(sessionId))
            : Model.withServerClock(this.clock.time);
        model.root = this.root.clone(model);
        model.api.batch = this.api.batch.clone();
        return model;
    };
    Model.prototype.clone = function () {
        return this.fork(this.clock.getSessionId());
    };
    Model.prototype.toView = function () {
        return this.root.toJson();
    };
    Model.prototype.toString = function (tab) {
        return this.root.toString(tab);
    };
    return Model;
}());
exports.Model = Model;
