"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Decoder = void 0;
var __1 = require("..");
var JsonPackExtension_1 = require("../JsonPackExtension");
var CachedKeyDecoder_1 = require("./CachedKeyDecoder");
var sharedCachedKeyDecoder = new CachedKeyDecoder_1.CachedKeyDecoder();
var Decoder = (function () {
    function Decoder(keyDecoder) {
        if (keyDecoder === void 0) { keyDecoder = sharedCachedKeyDecoder; }
        this.keyDecoder = keyDecoder;
        this.uint8 = new Uint8Array([]);
        this.view = new DataView(this.uint8.buffer);
        this.x = 0;
    }
    Decoder.prototype.getOffset = function () {
        return this.x;
    };
    Decoder.prototype.reset = function (uint8) {
        this.x = 0;
        this.uint8 = uint8;
        this.view = new DataView(uint8.buffer, uint8.byteOffset, uint8.byteLength);
    };
    Decoder.prototype.decode = function (uint8) {
        this.reset(uint8);
        return this.val();
    };
    Decoder.prototype.val = function () {
        var byte = this.u8();
        if (byte >= 0xe0)
            return byte - 0x100;
        if (byte <= 0xbf) {
            if (byte < 0x90) {
                if (byte <= 127)
                    return byte;
                return this.obj(byte & 15);
            }
            else {
                if (byte < 0xa0)
                    return this.arr(byte & 15);
                else
                    return this.str(byte & 31);
            }
        }
        if (byte <= 0xce) {
            if (byte <= 0xc8) {
                if (byte <= 0xc4) {
                    if (byte <= 0xc2)
                        return byte === 0xc2 ? false : null;
                    else
                        return byte === 0xc4 ? this.bin(this.u8()) : true;
                }
                else {
                    if (byte <= 0xc6)
                        return byte === 0xc6 ? this.bin(this.u32()) : this.bin(this.u16());
                    else
                        return byte === 0xc8 ? this.ext(this.u16()) : this.ext(this.u8());
                }
            }
            else {
                return byte <= 0xcc
                    ? byte <= 0xca
                        ? byte === 0xca
                            ? this.f32()
                            : this.ext(this.u32())
                        : byte === 0xcc
                            ? this.u8()
                            : this.f64()
                    : byte <= 0xce
                        ? byte === 0xce
                            ? this.u32()
                            : this.u16()
                        : byte === 0xd0
                            ? this.i8()
                            : this.u32() * 4294967296 + this.u32();
            }
        }
        else if (byte <= 0xd8) {
            return byte <= 0xd4
                ? byte <= 0xd2
                    ? byte === 0xd2
                        ? this.i32()
                        : this.i16()
                    : byte === 0xd4
                        ? this.ext(1)
                        : this.i32() * 4294967296 + this.i32()
                : byte <= 0xd6
                    ? byte === 0xd6
                        ? this.ext(4)
                        : this.ext(2)
                    : byte === 0xd8
                        ? this.ext(16)
                        : this.ext(8);
        }
        else {
            switch (byte) {
                case 0xd9:
                    return this.str(this.u8());
                case 0xda:
                    return this.str(this.u16());
                case 0xdb:
                    return this.str(this.u32());
                case 0xdc:
                    return this.arr(this.u16());
                case 0xdd:
                    return this.arr(this.u32());
                case 0xde:
                    return this.obj(this.u16());
                case 0xdf:
                    return this.obj(this.u32());
            }
        }
        return undefined;
    };
    Decoder.prototype.str = function (size) {
        var uint8 = this.uint8;
        var end = this.x + size;
        var x = this.x;
        var str = '';
        while (x < end) {
            var b1 = uint8[x++];
            if ((b1 & 0x80) === 0) {
                str += String.fromCharCode(b1);
                continue;
            }
            else if ((b1 & 0xe0) === 0xc0) {
                str += String.fromCharCode(((b1 & 0x1f) << 6) | (uint8[x++] & 0x3f));
            }
            else if ((b1 & 0xf0) === 0xe0) {
                str += String.fromCharCode(((b1 & 0x1f) << 12) | ((uint8[x++] & 0x3f) << 6) | (uint8[x++] & 0x3f));
            }
            else if ((b1 & 0xf8) === 0xf0) {
                var b2 = uint8[x++] & 0x3f;
                var b3 = uint8[x++] & 0x3f;
                var b4 = uint8[x++] & 0x3f;
                var code = ((b1 & 0x07) << 0x12) | (b2 << 0x0c) | (b3 << 0x06) | b4;
                if (code > 0xffff) {
                    code -= 0x10000;
                    str += String.fromCharCode(((code >>> 10) & 0x3ff) | 0xd800);
                    code = 0xdc00 | (code & 0x3ff);
                }
                str += String.fromCharCode(code);
            }
            else {
                str += String.fromCharCode(b1);
            }
        }
        this.x = end;
        return str;
    };
    Decoder.prototype.obj = function (size) {
        var obj = {};
        for (var i = 0; i < size; i++) {
            var key = this.key();
            obj[key] = this.val();
        }
        return obj;
    };
    Decoder.prototype.key = function () {
        var byte = this.view.getUint8(this.x);
        if (byte >= 160 && byte <= 191) {
            var size = byte & 31;
            var key = this.keyDecoder.decode(this.uint8, this.x + 1, size);
            this.x += 1 + size;
            return key;
        }
        else if (byte === 0xd9) {
            var size = this.view.getUint8(this.x + 1);
            if (size < 32) {
                var key = this.keyDecoder.decode(this.uint8, this.x + 2, size);
                this.x += 2 + size;
                return key;
            }
            else
                return this.val();
        }
        else
            return this.val();
    };
    Decoder.prototype.arr = function (size) {
        var arr = [];
        for (var i = 0; i < size; i++)
            arr.push(this.val());
        return arr;
    };
    Decoder.prototype.bin = function (size) {
        var end = this.x + size;
        var bin = this.uint8.subarray(this.x, end);
        this.x = end;
        return bin;
    };
    Decoder.prototype.ext = function (size) {
        var type = this.u8();
        var end = this.x + size;
        var buf = this.uint8.subarray(this.x, end);
        this.x = end;
        return new JsonPackExtension_1.JsonPackExtension(type, buf);
    };
    Decoder.prototype.back = function (bytes) {
        this.x -= bytes;
    };
    Decoder.prototype.peak = function () {
        return this.view.getUint8(this.x);
    };
    Decoder.prototype.u8 = function () {
        return this.view.getUint8(this.x++);
    };
    Decoder.prototype.u16 = function () {
        var num = this.view.getUint16(this.x);
        this.x += 2;
        return num;
    };
    Decoder.prototype.u32 = function () {
        var num = this.view.getUint32(this.x);
        this.x += 4;
        return num;
    };
    Decoder.prototype.i8 = function () {
        return this.view.getInt8(this.x++);
    };
    Decoder.prototype.i16 = function () {
        var num = this.view.getInt16(this.x);
        this.x += 2;
        return num;
    };
    Decoder.prototype.i32 = function () {
        var num = this.view.getInt32(this.x);
        this.x += 4;
        return num;
    };
    Decoder.prototype.f32 = function () {
        var pos = this.x;
        this.x += 4;
        return this.view.getFloat32(pos);
    };
    Decoder.prototype.f64 = function () {
        var pos = this.x;
        this.x += 8;
        return this.view.getFloat64(pos);
    };
    Decoder.prototype.decodeOneLevel = function (uint8) {
        this.reset(uint8);
        return this.valOneLevel();
    };
    Decoder.prototype.valOneLevel = function () {
        var byte = this.view.getUint8(this.x);
        var isMap = byte === 0xde || byte === 0xdf || byte >> 4 === 8;
        if (isMap) {
            this.x++;
            var size = byte === 0xde ? this.u16() : byte === 0xdf ? this.u32() : byte & 15;
            var obj = {};
            for (var i = 0; i < size; i++) {
                var key = this.key();
                obj[key] = this.primitive();
            }
            return obj;
        }
        var isArray = byte === 0xdc || byte === 0xdd || byte >> 4 === 9;
        if (isArray) {
            this.x++;
            var size = byte === 0xdc ? this.u16() : byte === 0xdd ? this.u32() : byte & 15;
            var arr = [];
            for (var i = 0; i < size; i++)
                arr.push(this.primitive());
            return arr;
        }
        return this.val();
    };
    Decoder.prototype.primitive = function () {
        var byte = this.view.getUint8(this.x);
        var isMapOrArray = byte === 0xde || byte === 0xdf || byte === 0xdc || byte === 0xdd || byte >> 5 === 4;
        if (isMapOrArray) {
            var length_1 = this.valSkip();
            this.x -= length_1;
            var buf = this.bin(length_1);
            return new __1.JsonPackValue(buf);
        }
        return this.val();
    };
    Decoder.prototype.skip = function (length) {
        this.x += length;
        return length;
    };
    Decoder.prototype.valSkip = function () {
        var byte = this.u8();
        if (byte >= 0xe0)
            return 1;
        if (byte <= 0xbf) {
            if (byte < 0x90) {
                if (byte <= 127)
                    return 1;
                return 1 + this.objSkip(byte & 15);
            }
            else {
                if (byte < 0xa0)
                    return 1 + this.arrSkip(byte & 15);
                else
                    return 1 + this.skip(byte & 31);
            }
        }
        if (byte <= 0xce) {
            if (byte <= 0xc8) {
                if (byte <= 0xc4) {
                    if (byte <= 0xc2)
                        return byte === 0xc2 ? 1 : 1;
                    else
                        return byte === 0xc4 ? 2 + this.skip(this.u8()) : 1;
                }
                else {
                    if (byte <= 0xc6)
                        return byte === 0xc6 ? 5 + this.skip(this.u32()) : 3 + this.skip(this.u16());
                    else
                        return byte === 0xc8 ? 4 + this.skip(this.u16()) : 3 + this.skip(this.u8());
                }
            }
            else {
                return byte <= 0xcc
                    ? byte <= 0xca
                        ? byte === 0xca
                            ? 1 + this.skip(4)
                            : 1 + 1 + 4 + this.skip(this.u32())
                        : byte === 0xcc
                            ? 1 + this.skip(1)
                            : 1 + this.skip(8)
                    : byte <= 0xce
                        ? byte === 0xce
                            ? 1 + this.skip(4)
                            : 1 + this.skip(2)
                        : byte === 0xd0
                            ? 1 + this.skip(1)
                            : 1 + this.skip(8);
            }
        }
        else if (byte <= 0xd8) {
            return byte <= 0xd4
                ? byte <= 0xd2
                    ? byte === 0xd2
                        ? 1 + this.skip(4)
                        : 1 + this.skip(2)
                    : byte === 0xd4
                        ? 1 + this.skip(2)
                        : 1 + this.skip(8)
                : byte <= 0xd6
                    ? byte === 0xd6
                        ? 1 + this.skip(5)
                        : 1 + this.skip(3)
                    : byte === 0xd8
                        ? 1 + this.skip(17)
                        : 1 + this.skip(9);
        }
        else {
            switch (byte) {
                case 0xd9:
                    return 2 + this.skip(this.u8());
                case 0xda:
                    return 3 + this.skip(this.u16());
                case 0xdb:
                    return 5 + this.skip(this.u32());
                case 0xdc:
                    return 3 + this.arrSkip(this.u16());
                case 0xdd:
                    return 5 + this.arrSkip(this.u32());
                case 0xde:
                    return 3 + this.objSkip(this.u16());
                case 0xdf:
                    return 5 + this.objSkip(this.u32());
            }
        }
        return 1;
    };
    Decoder.prototype.arrSkip = function (size) {
        var length = 0;
        for (var i = 0; i < size; i++)
            length += this.valSkip();
        return length;
    };
    Decoder.prototype.objSkip = function (size) {
        var length = 0;
        for (var i = 0; i < size; i++) {
            length += this.valSkip() + this.valSkip();
        }
        return length;
    };
    return Decoder;
}());
exports.Decoder = Decoder;
