"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTypeSystem = void 0;
var tslib_1 = require("tslib");
var json_type_validator_1 = require("../json-type-validator");
var dynamicFunction_1 = require("../util/codegen/dynamicFunction");
var json_type_serializer_1 = require("../json-type-serializer");
var util_1 = require("../json-pack/util");
var json_type_schema_1 = require("../json-type-schema");
var toTypeScriptText_1 = require("../json-type-typescript/toTypeScriptText");
var export_1 = require("../json-type-typescript/export");
var literalTypes = ['bin', 'bool', 'nil', 'num', 'str'];
var JsonTypeSystem = (function () {
    function JsonTypeSystem(options) {
        var _this = this;
        this.options = options;
        this.ref = function (ref) {
            var type = _this.options.types[ref];
            if (!type)
                throw new Error("Type [ref = ".concat(ref, "] not found."));
            return type;
        };
        this.fastValidatorCache = {};
        this.fastValidatorUsage = {};
        this.fullValidatorCache = {};
        this.jsonSerializerCache = {};
        this.jsonSerializerUsage = {};
        this.msgPackPartialEncoderCache = {};
        this.msgPackPartialEncoderUsage = {};
        this.msgPackEncoderCache = {};
    }
    JsonTypeSystem.prototype.hasType = function (ref) {
        return Object.prototype.hasOwnProperty.call(this.options.types, ref);
    };
    JsonTypeSystem.prototype.getFastValidator = function (ref) {
        var _this = this;
        var type = this.ref(ref);
        var cachedFastValidator = this.fastValidatorCache[ref];
        if (cachedFastValidator) {
            this.fastValidatorUsage[ref] = (this.fastValidatorUsage[ref] || 0) + 1;
            return cachedFastValidator;
        }
        var _a = (0, tslib_1.__read)((0, dynamicFunction_1.dynamicFunction)(function () {
            throw new Error("Type [ref = ".concat(ref, "] not implemented."));
        }), 2), validator = _a[0], setValidator = _a[1];
        this.fastValidatorCache[ref] = validator;
        var realValidator = (0, json_type_validator_1.createBoolValidator)(type, {
            customValidators: this.options.customValidators,
            ref: function (id) { return _this.getFastValidator(id); },
            skipObjectExtraFieldsCheck: true,
            unsafeMode: true,
        });
        if (this.fastValidatorUsage[ref])
            setValidator(realValidator);
        this.fastValidatorCache[ref] = realValidator;
        return realValidator;
    };
    JsonTypeSystem.prototype.getFullValidator = function (ref) {
        var _this = this;
        var type = this.ref(ref);
        var cachedFullValidator = this.fullValidatorCache[ref];
        if (cachedFullValidator)
            return cachedFullValidator;
        var _a = (0, tslib_1.__read)((0, dynamicFunction_1.dynamicFunction)(function () {
            throw new Error("Type [ref = ".concat(ref, "] not implemented."));
        }), 2), validator = _a[0], setValidator = _a[1];
        this.fullValidatorCache[ref] = validator;
        setValidator((0, json_type_validator_1.createObjValidator)(type, {
            customValidators: this.options.customValidators,
            ref: function (id) { return _this.getFullValidator(id); },
        }));
        return validator;
    };
    JsonTypeSystem.prototype.getJsonSerializer = function (ref) {
        var _this = this;
        var type = this.ref(ref);
        var jsonSerializerCached = this.jsonSerializerCache[ref];
        if (jsonSerializerCached) {
            this.jsonSerializerUsage[ref] = (this.jsonSerializerUsage[ref] || 0) + 1;
            return jsonSerializerCached;
        }
        var codegen = new json_type_serializer_1.JsonSerializerCodegen({
            type: type,
            ref: function (id) {
                var type = _this.ref(id);
                var isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : _this.getJsonSerializer(id);
            },
        });
        var _a = (0, tslib_1.__read)((0, dynamicFunction_1.dynamicFunction)(function () {
            throw new Error("Type [ref = ".concat(ref, "] JSON serializer not implemented."));
        }), 2), serializer = _a[0], setSerializer = _a[1];
        this.jsonSerializerCache[ref] = serializer;
        var realSerializer = codegen.run().compile();
        if (this.jsonSerializerUsage[ref])
            setSerializer(realSerializer);
        this.jsonSerializerCache[ref] = realSerializer;
        return realSerializer;
    };
    JsonTypeSystem.prototype.getMsgPackPartialEncoder = function (ref) {
        var _this = this;
        var type = this.ref(ref);
        var msgPackPartialEncoderCached = this.msgPackPartialEncoderCache[ref];
        if (msgPackPartialEncoderCached) {
            this.msgPackPartialEncoderUsage[ref] = (this.msgPackPartialEncoderUsage[ref] || 0) + 1;
            return msgPackPartialEncoderCached;
        }
        var codegen = new json_type_serializer_1.MsgPackSerializerCodegen({
            type: type,
            encoder: util_1.encoderFull,
            ref: function (id) {
                var type = _this.ref(id);
                var isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : _this.getMsgPackPartialEncoder(id);
            },
        });
        var _a = (0, tslib_1.__read)((0, dynamicFunction_1.dynamicFunction)(function () {
            throw new Error("Type [ref = ".concat(ref, "] MessagePack partial encoder not implemented."));
        }), 2), partialEncoder = _a[0], setPartialEncoder = _a[1];
        this.msgPackPartialEncoderCache[ref] = partialEncoder;
        var realPartialEncoder = codegen.run().compilePartial();
        if (this.msgPackPartialEncoderUsage[ref])
            setPartialEncoder(realPartialEncoder);
        this.msgPackPartialEncoderCache[ref] = realPartialEncoder;
        return realPartialEncoder;
    };
    JsonTypeSystem.prototype.getMsgPackEncoder = function (ref) {
        var _this = this;
        var type = this.ref(ref);
        var msgPackSerializerCached = this.msgPackEncoderCache[ref];
        if (msgPackSerializerCached)
            return msgPackSerializerCached;
        var codegen = new json_type_serializer_1.MsgPackSerializerCodegen({
            type: type,
            encoder: util_1.encoderFull,
            ref: function (id) {
                var type = _this.ref(id);
                var isLiteralType = literalTypes.indexOf(type.__t) !== -1;
                return isLiteralType ? type : _this.getMsgPackPartialEncoder(id);
            },
        });
        return (this.msgPackEncoderCache[ref] = codegen.run().compile());
    };
    JsonTypeSystem.prototype.toJsonSchema = function (ref, withDefs) {
        var type = this.ref(ref);
        var defs = withDefs ? {} : undefined;
        var schema = (0, json_type_schema_1.toJsonSchema)(type, {
            ref: this.ref,
            defs: defs,
        });
        if (defs) {
            schema = (0, tslib_1.__assign)((0, tslib_1.__assign)({ $id: ref }, schema), { $defs: defs });
        }
        return schema;
    };
    JsonTypeSystem.prototype.toTypeScriptText = function (ref) {
        var type = this.ref(ref);
        return (0, toTypeScriptText_1.toTypeScriptText)(type, '');
    };
    JsonTypeSystem.prototype.toTsAst = function (ref) {
        var e_1, _a;
        var statements = [];
        if (ref) {
            (0, export_1.exportDeclaration)(ref, {
                ref: this.ref,
                statements: statements,
            });
        }
        else {
            var types = this.options.types;
            try {
                for (var _b = (0, tslib_1.__values)(Object.keys(types)), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var ref_1 = _c.value;
                    (0, export_1.exportDeclaration)(ref_1, {
                        ref: this.ref,
                        statements: statements,
                    });
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                }
                finally { if (e_1) throw e_1.error; }
            }
        }
        return statements;
    };
    return JsonTypeSystem;
}());
exports.JsonTypeSystem = JsonTypeSystem;
