"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RpcServer = exports.RpcServerError = void 0;
var tslib_1 = require("tslib");
var nominal_1 = require("../messages/nominal");
var subscribeCompleteObserver_1 = require("../util/subscribeCompleteObserver");
var TimedQueue_1 = require("../util/TimedQueue");
var constants_1 = require("./constants");
Object.defineProperty(exports, "RpcServerError", { enumerable: true, get: function () { return constants_1.RpcServerError; } });
var error_1 = require("./error");
var RpcServer = (function () {
    function RpcServer(_a) {
        var _this = this;
        var caller = _a.caller, error = _a.error, send = _a.send, notify = _a.onNotification, _b = _a.bufferSize, bufferSize = _b === void 0 ? 10 : _b, _c = _a.bufferTime, bufferTime = _c === void 0 ? 1 : _c;
        this.activeStreamCalls = new Map();
        this.onStreamError = function (id, error) {
            _this.send(_this.resErrorMessage(id, error));
            _this.activeStreamCalls.delete(id);
        };
        this.caller = caller;
        this.error = error || new error_1.ErrorLikeErrorFormatter();
        this.onNotification = notify;
        this.onSend = send;
        if (bufferTime) {
            var buffer_1 = new TimedQueue_1.TimedQueue();
            buffer_1.itemLimit = bufferSize;
            buffer_1.timeLimit = bufferTime;
            buffer_1.onFlush = function (messages) { return _this.onSend(messages); };
            this.send = function (message) {
                buffer_1.push(message);
            };
        }
        else {
            this.send = function (message) {
                _this.onSend([message]);
            };
        }
    }
    RpcServer.prototype.notifMessage = function (method, data) {
        return new nominal_1.NotificationMessage(method, data);
    };
    RpcServer.prototype.resCompleteMessage = function (id, data) {
        return new nominal_1.ResponseCompleteMessage(id, data);
    };
    RpcServer.prototype.resDataMessage = function (id, data) {
        return new nominal_1.ResponseDataMessage(id, data);
    };
    RpcServer.prototype.resErrorMessage = function (id, data) {
        return new nominal_1.ResponseErrorMessage(id, data);
    };
    RpcServer.prototype.reqUnsubscribeMessage = function (id) {
        return new nominal_1.RequestUnsubscribeMessage(id);
    };
    RpcServer.prototype.createStaticCall = function (name, request, ctx) {
        return this.caller.call(name, request, ctx);
    };
    RpcServer.prototype.execStaticCall = function (id, name, request, ctx) {
        var _this = this;
        this.createStaticCall(name, request, ctx)
            .then(function (response) {
            _this.send(_this.resCompleteMessage(id, response));
        })
            .catch(function (error) {
            _this.send(_this.resErrorMessage(id, error));
        });
    };
    RpcServer.prototype.createCall = function (name, ctx) {
        var _this = this;
        var streamCallback = function (method, ctx, req$) { return method.call$(ctx, req$); };
        return this.caller.createCall(name, ctx, function (name, req, ctx) { return _this.caller.call(name, req, ctx); }, streamCallback);
    };
    RpcServer.prototype.onMessage = function (message, ctx) {
        try {
            if (message instanceof nominal_1.RequestDataMessage)
                this.onRequestDataMessage(message, ctx);
            else if (message instanceof nominal_1.RequestCompleteMessage)
                this.onRequestCompleteMessage(message, ctx);
            else if (message instanceof nominal_1.RequestErrorMessage)
                this.onRequestErrorMessage(message, ctx);
            else if (message instanceof nominal_1.NotificationMessage)
                this.onNotificationMessage(message, ctx);
            else if (message instanceof nominal_1.ResponseUnsubscribeMessage)
                this.onUnsubscribeMessage(message);
        }
        catch (error) {
            var formattedError = this.error.format(error);
            var message_1 = this.notifMessage('.err', formattedError);
            this.send(message_1);
        }
    };
    RpcServer.prototype.onMessages = function (messages, ctx) {
        var length = messages.length;
        for (var i = 0; i < length; i++)
            this.onMessage(messages[i], ctx);
    };
    RpcServer.prototype.stop = function (reason) {
        var e_1, _a;
        if (reason === void 0) { reason = constants_1.RpcServerError.Stop; }
        this.send = (function () { });
        this.onNotification = function (name, data, ctx) { };
        try {
            for (var _b = (0, tslib_1.__values)(this.activeStreamCalls.values()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var call = _c.value;
                call.req$.error(new error_1.RpcError(reason));
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.activeStreamCalls.clear();
    };
    RpcServer.prototype.disconnect = function () {
        this.stop(constants_1.RpcServerError.Disconnect);
    };
    RpcServer.prototype.sendError = function (id, code) {
        var data = this.error.formatCode(code);
        var message = this.resErrorMessage(id, data);
        this.send(message);
    };
    RpcServer.prototype.createStreamCall = function (id, name, ctx) {
        var _this = this;
        var call = this.createCall(name, ctx);
        this.activeStreamCalls.set(id, call);
        (0, subscribeCompleteObserver_1.subscribeCompleteObserver)(call.res$, {
            next: function (value) {
                _this.send(_this.resDataMessage(id, value));
            },
            error: function (error) {
                _this.onStreamError(id, error);
            },
            complete: function (value) {
                _this.send(_this.resCompleteMessage(id, value));
                _this.activeStreamCalls.delete(id);
            },
        });
        call.reqUnsubscribe$.subscribe(function () {
            if (_this.activeStreamCalls.has(id))
                _this.send(_this.reqUnsubscribeMessage(id));
        });
        return call;
    };
    RpcServer.prototype.onRequestDataMessage = function (message, ctx) {
        var id = message.id, method = message.method, data = message.data;
        var call = this.activeStreamCalls.get(id);
        if (!call) {
            if (!method) {
                this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
                return;
            }
            call = this.createStreamCall(id, method, ctx);
        }
        if (call) {
            if (data !== undefined) {
                call.req$.next(data);
            }
        }
    };
    RpcServer.prototype.onRequestCompleteMessage = function (message, ctx) {
        var id = message.id, method = message.method, data = message.data;
        var call = this.activeStreamCalls.get(id);
        if (call) {
            var req$ = call.req$;
            if (data !== undefined)
                req$.next(data);
            req$.complete();
            return;
        }
        if (!method) {
            this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
            return;
        }
        if (!this.caller.exists(method)) {
            this.sendError(id, constants_1.RpcServerError.MethodNotFound);
            return;
        }
        var rpcMethod = this.caller.get(method);
        if (rpcMethod.isStreaming) {
            var newCall = this.createStreamCall(id, method, ctx);
            if (newCall) {
                if (data !== undefined) {
                    newCall.req$.next(data);
                    newCall.req$.complete();
                }
            }
        }
        else
            this.execStaticCall(id, method, data, ctx);
    };
    RpcServer.prototype.onRequestErrorMessage = function (message, ctx) {
        var id = message.id, method = message.method, data = message.data;
        var call = this.activeStreamCalls.get(id);
        if (call) {
            return call.req$.error(data);
        }
        if (!method)
            return this.sendError(id, constants_1.RpcServerError.NoMethodSpecified);
        if (!this.caller.exists(method))
            return this.sendError(id, constants_1.RpcServerError.MethodNotFound);
        var rpcMethod = this.caller.get(method);
        if (!rpcMethod.isStreaming)
            return this.sendError(id, constants_1.RpcServerError.ErrorForStaticMethod);
        var streamCall = this.createStreamCall(id, method, ctx);
        if (!streamCall)
            return;
        streamCall.req$.error(data);
    };
    RpcServer.prototype.onUnsubscribeMessage = function (message) {
        var id = message.id;
        var call = this.activeStreamCalls.get(id);
        if (!call)
            return;
        this.activeStreamCalls.delete(id);
        call.req$.complete();
    };
    RpcServer.prototype.onNotificationMessage = function (message, ctx) {
        var method = message.method, data = message.data;
        if (!method || method.length > 128)
            throw new error_1.RpcError(constants_1.RpcServerError.InvalidNotificationName);
        this.onNotification(method, data, ctx);
    };
    return RpcServer;
}());
exports.RpcServer = RpcServer;
